(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Argument_Filter
imports
  Term_Order_Impl
  Trs_Impl
  Map_Choice
  Relative_Rewriting
begin

datatype_new af_entry = Collapse nat | AFList "nat list"
type_synonym 'f afs = "'f \<times> nat \<Rightarrow> af_entry"
datatype_new ('f) filtered = FPair 'f  nat
datatype_compat filtered

(* the main issue to not use pairs is that we can provide an alternative show function *)

derive linorder filtered
instantiation filtered :: (key)key
begin
instance ..
end

fun
  filtered_fun :: "'f filtered \<Rightarrow> 'f"
where
  "filtered_fun (FPair f n) = f"

instantiation filtered :: ("show") "show" begin
  definition "shows_prec (d::nat) ff = shows (filtered_fun ff)"
  lemma assoc_filtered: "shows_prec d (p::'a filtered) r @ s = shows_prec d p (r @ s)"
    by (simp add: shows_prec_filtered_def)
  standard_shows_list assoc_filtered
end

fun apply_af_entry :: "'f \<Rightarrow> af_entry \<Rightarrow> ('f,'v)term list \<Rightarrow> ('f,'v)term" where
  "apply_af_entry _ (Collapse i) ts = ts ! i"
| "apply_af_entry f (AFList is) ts = Fun f (map (\<lambda> i. ts ! i) is)"

fun wf_af_entry :: "nat \<Rightarrow> af_entry \<Rightarrow> bool"
where "wf_af_entry n (Collapse i) = (i < n)"
  |   "wf_af_entry n (AFList is) = Ball (set is) (\<lambda> i. i < n)"

definition wf_afs :: "'f afs \<Rightarrow> bool"
where "wf_afs \<pi> \<equiv> \<forall> f n. wf_af_entry n (\<pi> (f,n))"

fun afs_term :: "'f afs \<Rightarrow> ('f,'v)term \<Rightarrow> ('f filtered,'v)term" where 
  "afs_term \<pi> (Fun f ts) = (let l = length ts in apply_af_entry (FPair f l) (\<pi> (f,l)) (map (afs_term \<pi>) ts))"
| "afs_term \<pi> (Var x) = Var x"

definition af_term :: "'f afs \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term" where
  "af_term \<pi> t = map_term filtered_fun (afs_term \<pi> t)"

definition afs_rule :: "'f afs \<Rightarrow> ('f,'v)rule \<Rightarrow> ('f filtered,'v)rule"
where "afs_rule \<pi> lr \<equiv> (afs_term \<pi> (fst lr), afs_term \<pi> (snd lr))"

definition af_rule :: "'f afs \<Rightarrow> ('f,'v)rule \<Rightarrow> ('f,'v)rule" where
  "af_rule \<pi> \<equiv> \<lambda> t. map_rule filtered_fun (afs_rule \<pi> t)"

definition afs_rules :: "'f afs \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f filtered,'v)rules"
where "afs_rules \<pi> R = map (afs_rule \<pi>) R"

definition af_rules :: "'f afs \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rules"
where "af_rules \<pi> R = map (af_rule \<pi>) R"

definition
  afs_check ::
    "shows \<Rightarrow> 'f::show afs \<Rightarrow> (('f filtered,'v::show)rule \<Rightarrow> shows check) \<Rightarrow> ('f,'v)rule  \<Rightarrow> shows check"
where
  "afs_check r \<pi> g lr \<equiv> let pl = afs_term \<pi> (fst lr); pr = afs_term \<pi> (snd lr) in g (pl,pr) <+? (\<lambda>s. 
  ''could not orient '' +#+ shows(fst lr) +@+ '' '' +#+ r +@+ '' '' +#+ shows(snd lr) +@+ shows_nl +@+
  ''pi( '' +#+ shows(fst lr) +@+ '' ) = '' +#+ shows pl +@+ shows_nl +@+
  ''pi( '' +#+ shows(snd lr) +@+ '' ) = '' +#+ shows pr +@+ shows_nl +@+ s
)"

lemma af_rules[simp]: "set (af_rules \<pi> R) = af_rule \<pi> ` (set R)" unfolding af_rules_def by auto
 
locale afs =
  fixes \<pi> :: "'f afs"
  assumes wf_afs: "wf_afs \<pi>"
begin

abbreviation afs_subst :: "('f,'v)subst \<Rightarrow> ('f filtered,'v)subst"
where "afs_subst \<sigma> \<equiv> (\<lambda>x. afs_term \<pi> (\<sigma> x))"

definition afs_rel :: "('f filtered,'v)term rel \<Rightarrow> ('f,'v)term rel"
where "afs_rel R = {lr | lr. afs_rule \<pi> lr \<in> R}"

lemma afs_subst[simp]: "afs_term \<pi> (t \<cdot> \<sigma>) = afs_term \<pi> t \<cdot> afs_subst \<sigma>"
proof -
  have "afs_term \<pi> (t \<cdot> \<sigma>) = afs_term \<pi> t \<cdot> afs_subst (\<sigma>)"
  proof (induct t)
    case (Fun f ss)
    let ?l = "length ss"
    let ?e = "\<pi> (f,?l)"
    from wf_afs have wf_e: "wf_af_entry ?l ?e" unfolding wf_afs_def by auto
    show ?case
    proof (cases ?e)
      case (Collapse i)
      with wf_e have "i < ?l" by simp
      with Collapse Fun(1)[of "ss ! i"] show ?thesis
	by (auto simp: Let_def)
    next
      case (AFList "is")
      with wf_e have wf: "\<forall> i \<in> set is. i < ?l" by auto
      with AFList Fun(1) show ?thesis 
	by auto
    qed
  qed simp
  thus ?thesis by simp
qed


lemma afs_rsteps:
  "(s,t) \<in> (rstep R)^* \<Longrightarrow> (afs_term \<pi> s, afs_term \<pi> t) \<in> (rstep (afs_rule \<pi> ` R))^*"
proof (induct rule: rtrancl_induct)
  case (step t u)
  let ?a = "afs_term \<pi>"
  let ?R = "afs_rule \<pi> ` R"
  let ?RR = "(rstep ?R)^*"
  from rstepE[OF step(2)] obtain C l r \<sigma> where t: "t = C \<langle> l \<cdot> \<sigma> \<rangle>" and u: "u = C \<langle> r \<cdot> \<sigma> \<rangle>" and lr: "(l,r) \<in> R" .
  def ll \<equiv> "l \<cdot> \<sigma>" def rr \<equiv> "r \<cdot> \<sigma>"
  note ctxt = all_ctxt_closedE[OF rsteps_all_ctxt_closed[of _ ?R], of _ _ UNIV]
  have "(?a t, ?a u) \<in> ?RR" unfolding t u ll_def[symmetric] rr_def[symmetric]
  proof (induct C)
    case Hole
    from lr have "(?a l, ?a r) \<in> ?R" by (force simp: afs_rule_def)
    from rstep.subst[OF rstep.id[OF this], of "afs_subst \<sigma>"]
    show ?case unfolding ll_def rr_def by simp
  next
    case (More f bef C aft)
    let ?n = "Suc (length bef + length aft)"    
    let ?pi = "\<pi> (f, ?n)"
    show ?case
    proof (cases ?pi)
      case (Collapse i)
      thus ?thesis 
        by (cases "i = length bef", insert More, auto simp: nth_append)
    next
      case (AFList ls)
      show ?thesis 
        by (simp add: AFList, rule ctxt, auto, case_tac "ls ! i = length bef", insert More, auto simp: nth_append)
    qed
  qed
  with step(3) show ?case by simp
qed simp

lemma af_rsteps: assumes steps: "(s,t) \<in> (rstep R)^*"
  shows "(af_term \<pi> s, af_term \<pi> t) \<in> (rstep (af_rule \<pi> ` R))^*"
proof -
  let ?f = "filtered_fun"
  have id: "map_trs ?f (afs_rule \<pi> ` R) = af_rule \<pi> ` R"
    by (force simp: af_rule_def afs_rule_def af_term_def map_trs.simps)
  from rsteps_imp_map_rsteps[OF afs_rsteps[OF steps], of ?f, folded af_term_def]
  show ?thesis unfolding id .
qed

lemma argument_filter_nj: "\<not> (\<exists> u. (af_term \<pi> s,u) \<in> (rstep (af_rule \<pi> ` Rs))^* \<and> (af_term \<pi> t,u) \<in> (rstep (af_rule \<pi> ` Rt))^*) \<Longrightarrow>
  \<not> (\<exists> u. (s,u) \<in> (rstep Rs)^* \<and> (t,u) \<in> (rstep Rt)^*)"
  using af_rsteps by auto

lemma afs_subst_closed: 
  assumes "subst_closed R"
  shows "subst_closed (afs_rel R)"
unfolding subst_closed_def
proof (rule subsetI, simp only: split_paired_all)
  fix ss ts
  assume "(ss,ts) \<in> CS (afs_rel R)"
  thus "(ss,ts) \<in> afs_rel R"
  proof (induct)
    case (subst_closureI s t \<sigma>)
    from this have "(afs_term \<pi> s, afs_term \<pi> t) \<in> R" unfolding afs_rel_def afs_rule_def by simp
    with assms have "(afs_term \<pi> s \<cdot> afs_subst \<sigma>, afs_term \<pi> t \<cdot> afs_subst \<sigma>) \<in> R" unfolding subst_closed_def 
      using subst_closure.subst_closureI by auto    
    thus ?case unfolding afs_rel_def afs_rule_def by auto
  qed
qed

lemma afs_SN: 
  assumes "SN R"
  shows "SN (afs_rel R)"
proof -
  have id: "afs_rel R = inv_image R (afs_term \<pi>)" 
    unfolding afs_rel_def afs_rule_def inv_image_def by auto
  show ?thesis unfolding id by (rule SN_inv_image[OF assms])
qed

lemma afs_compat: 
  assumes "NS O S \<subseteq> S"
  shows "afs_rel NS O afs_rel S \<subseteq> afs_rel S"
using assms
  by (unfold afs_rel_def afs_rule_def, clarify, auto)

lemma afs_compat2: 
  assumes "S O NS \<subseteq> S"
  shows "afs_rel S O afs_rel NS \<subseteq> afs_rel S"
using assms
  by (unfold afs_rel_def afs_rule_def, clarify, auto)

lemma afs_refl: 
  assumes "refl NS"
  shows "refl (afs_rel NS)"
using assms unfolding refl_on_def
  by (unfold afs_rel_def afs_rule_def, auto)

lemma afs_trans: 
  assumes "trans r"
  shows "trans (afs_rel r)"
unfolding trans_def
proof (unfold afs_rel_def afs_rule_def, clarify, unfold fst_conv snd_conv)
  fix a b c
  assume gt: "(afs_term \<pi> a, afs_term \<pi> b) \<in> r" "(afs_term \<pi> b, afs_term \<pi> c) \<in> r"
  show "\<exists> lr. (a,c) = lr \<and> (afs_term \<pi> (fst lr), afs_term \<pi> (snd lr)) \<in> r"
    by (intro exI conjI, rule refl, insert assms[unfolded trans_def, THEN spec, THEN spec, THEN spec, THEN mp[OF _ gt(1)], THEN mp[OF _ gt(2)]], simp)
qed

lemma afs_NS_mono: 
  assumes acc: "all_ctxt_closed UNIV R"
  shows "ctxt_closed (afs_rel R)"
proof (rule one_imp_ctxt_closed)
  fix f bef s t aft
  assume st: "(s,t) \<in> afs_rel R"
  let ?ps = "afs_term \<pi> s"
  let ?pt = "afs_term \<pi> t"
  let ?bsa = "bef @ s # aft"
  let ?bta = "bef @ t # aft"
  from st have pst: "(?ps,?pt) \<in> R" unfolding afs_rel_def afs_rule_def by auto
  let ?l = "Suc (length bef + length aft)"
  let ?ls = "length (bef @ s # aft)"
  let ?lt = "length (bef @ t # aft)"
  let ?af = "\<pi> (f,?l)"
  from wf_afs have wf_af: "wf_af_entry ?l ?af" unfolding wf_afs_def by simp
  let ?msa = "map (afs_term \<pi>) ?bsa"
  let ?mta = "map (afs_term \<pi>) ?bta"
  let ?pfs = "apply_af_entry (FPair f ?l) ?af ?msa"
  let ?pft = "apply_af_entry (FPair f ?l) ?af ?mta"
  have all2: "\<forall> i < ?l. (?msa ! i, ?mta ! i) \<in> R"
  proof (intro allI impI)
    fix i
    assume "i < ?l"
    hence i_s: "i < ?ls" and i_t: "i < ?lt" by simp_all
    {
      assume bef: "i < length bef"
      with all_ctxt_closed_reflE[OF acc] have "(?msa ! i, ?mta ! i) \<in> R"  
        by (simp only: nth_map[OF i_s], simp add: nth_append)
    }
    moreover
    {
      assume "i = length bef"
      with pst have "(?msa ! i, ?mta ! i) \<in> R" unfolding refl_on_def
       	by (simp only: nth_map[OF i_s], simp add: nth_append)
    }
    moreover
    {
      assume i: "i > length bef"
      from this obtain k where k1: "i - length bef = Suc k" by (cases "i - length bef", arith+)
      with i_s have k2: "k < length aft" by auto
      from i all_ctxt_closed_reflE[OF acc] have "(?msa ! i, ?mta ! i) \<in> R"
        by (simp only: nth_map[OF i_s] nth_map[OF i_t], simp add: nth_append k1)
    }
    moreover
    {
      have "i < length bef \<or> i = length bef \<or> i > length bef" by auto
    }      
    ultimately
    show "(?msa ! i, ?mta ! i) \<in> R" by blast
  qed
  have "(?pfs,?pft) \<in> R"
  proof (cases ?af)
    case (Collapse i)
    with wf_af have "i < ?l"  by simp
    with all2 Collapse show ?thesis by auto
  next
    case (AFList ids)
    with wf_af have ids: "\<forall> i \<in> set ids. i < ?l" by simp
    show ?thesis 
    proof (simp only: AFList apply_af_entry.simps)
      let ?msi = "map (op ! (map (afs_term \<pi>) ?bsa)) ids"
      let ?mti = "map (op ! (map (afs_term \<pi>) ?bta)) ids"
      from ids all2 have args: "\<forall> i < length ids. (?msi ! i, ?mti ! i) \<in> R" by auto
      show "(Fun (FPair f ?l) ?msi, Fun (FPair f ?l) ?mti) \<in> R"
        by (rule all_ctxt_closedE[OF acc], insert args, auto)
    qed
  qed
  thus "(Fun f ?bsa, Fun f ?bta) \<in> afs_rel R" unfolding afs_rel_def afs_rule_def by simp
qed
end  

locale afs_redtriple = afs \<pi> + redtriple_order S NS NST
  for \<pi> :: "('f)afs" and S NS NST :: "('f filtered,'v)trs" 

locale afs_redtriple_impl = afs_redtriple \<pi> S NS NST + redtriple_impl S NS NST cS cNS cNST
  for \<pi> :: "('f::show)afs" and S NS NST :: "('f filtered,'v::show)trs" and cS cNS cNST


locale mono_afs_redtriple_impl = afs_redtriple_impl \<pi> S NS NST cS cNS cNST + mono_redtriple_impl S NS NST cS cNS cNST cm
  for \<pi> :: "('f::show)afs" and S NS NST :: "('f filtered,'v::show)trs" and cS cNS cNST cm

sublocale afs_redtriple \<subseteq> redtriple "afs_rel S" "afs_rel NS" "afs_rel NST"
proof -
  let ?S = "afs_rel S"
  let ?NS = "afs_rel NS"
  let ?NST = "afs_rel NST"
  show "redtriple ?S ?NS ?NST"
  proof
    show "SN ?S" 
      by (rule afs_SN, rule SN)
  next
    show "?NS O ?S \<subseteq> ?S"
      by (rule afs_compat, rule compat_NS_S)
  next
    show "?S O ?NS \<subseteq> ?S"
      by (rule afs_compat2, rule compat_S_NS)
  next
    show "subst_closed ?S"
      by (rule afs_subst_closed, rule subst_S)
  next
    show "subst_closed ?NS"
      by (rule afs_subst_closed, rule subst_NS)
  next
    show "subst_closed ?NST"
      by (rule afs_subst_closed, rule subst_NST)
  next
    show "ctxt_closed ?NS"
      by (rule afs_NS_mono, rule all_ctxt_closed)
  next
    show "?NST O ?S \<subseteq> ?S"
      by (rule afs_compat[OF compat_NST])
  qed
qed

fun mono_af_entry :: "nat \<Rightarrow> af_entry \<Rightarrow> bool"
where "mono_af_entry n (Collapse i) = (n \<le> 1)"
  | "mono_af_entry n (AFList ids) = Ball (set [0 ..< n]) (\<lambda> i. i \<in> set ids)"

definition mono_afs :: "'f afs \<Rightarrow> bool"
where "mono_afs \<pi> \<equiv> \<forall> f n. mono_af_entry n (\<pi> (f,n))"

context afs_redtriple
begin

lemma ctxt_closed: assumes mono_pi: "mono_afs \<pi>"
  and monoS: "ctxt_closed S"
  shows "ctxt_closed (afs_rel S)"
proof -
  let ?S = "afs_rel S"
  let ?pi = "afs_term \<pi>"
  {
    fix f bef s t aft
    assume st: "(?pi s, ?pi t) \<in> S"
    let ?bsa = "bef @ s # aft"
    let ?bta = "bef @ t # aft"
    let ?fs = "Fun f ?bsa"
    let ?ft = "Fun f ?bta"
    let ?l = "Suc (length bef + length aft)"
    let ?lba = "length bef + length aft"
    let ?e = "\<pi> (f,?l)"
    from mono_pi have mono_e: "mono_af_entry ?l ?e" unfolding mono_afs_def by auto
    from wf_afs have wf_e: "wf_af_entry ?l ?e" unfolding wf_afs_def by auto
    have "(?pi ?fs, ?pi ?ft) \<in> S"
    proof (cases ?e)
      case (Collapse i)
      with mono_e wf_e show ?thesis using st by auto
    next
      case (AFList ids)
      thus ?thesis 
      proof simp
	let ?ids = "{n | n. n < ?l}"
	let ?f = "FPair f ?l"
	obtain l where l: "?l = l" by auto
	from AFList wf_e have ids1: "set ids \<subseteq> ?ids" by auto
	from AFList mono_e have ids2: "?ids \<subseteq> set ids" 	
	  by (simp only: l, auto)
	from ids1 ids2 have ids: "set ids = ?ids" by auto
	let ?eq = "\<lambda> i. ?pi (?bsa ! i) = ?pi (?bta ! i)"
	let ?gr = "\<lambda> i. (?pi (?bsa ! i), ?pi (?bta ! i)) \<in> S"
	let ?lb = "length bef"
	have geq: "\<forall> i. ?eq i \<or> ?gr i"
	proof
	  fix i
	  {
	    assume "i < ?lb" hence "?eq i" by (simp add: nth_append)
	  }
	  moreover
	  {
	    assume "i = ?lb"
	    with st have "?gr i" by (simp add: nth_append)
	  }
	  moreover 
	  {
	    assume i2: "i > ?lb"
	    then obtain k where k1: "i - length bef = Suc k" by (cases "i - length bef", arith+)
	    have "?eq i" unfolding nth_append using i2 k1 by auto
	  }
	  moreover
	  {
	    have "i < ?lb \<or> i = ?lb \<or> i > ?lb" by arith
	  }
	  ultimately show "?eq i \<or> ?gr i" by blast
	qed
	from ids have lb: "?lb \<in> set ids" by auto
	from st have "?gr ?lb" by (simp add: nth_append)
	with lb have gr: "\<exists> i \<in> set ids. ?gr i \<and> i = ?lb" by auto
	obtain ss ts pi g where g: "g = ?f" and ss: "ss = ?bsa" and ts: "ts = ?bta" and pi: "pi = (?pi :: ('f,'v)term \<Rightarrow> ('f filtered,'v)term)" by auto
	let ?eq = "\<lambda> i. pi (ss ! i) = pi (ts ! i)"
	let ?gr = "\<lambda> i. (pi (ss ! i), pi (ts ! i)) \<in> S"
	let ?ss = "map (op ! (map pi ss))"
	let ?ts = "map (op ! (map pi ts))"
	let ?both = "\<lambda> n. ?ts (take n ids) @ ?ss (drop n ids)"
	let ?gss = "Fun g (?ss ids)"
	let ?gboth = "\<lambda> n. Fun g (?both n)"
	from geq ss ts pi have geq: "\<forall> i. ?eq i \<or> ?gr i" by auto
	from gr ss ts pi have "\<exists> i \<in> set ids. ?gr i \<and> i = ?lb" by auto
	from this obtain i where i: "?gr i" and iids: "i \<in> set ids" and idef: "i = ?lb" by auto
	have "\<forall> n. (i \<in> set (take n ids) \<longrightarrow> (?gss, ?gboth n) \<in> S) \<and> (i \<notin> set (take n ids) \<longrightarrow> ?gss = ?gboth n)" (is "\<forall> n. ?p n") 
	proof (intro allI)
	  fix n
	  show "?p n"
	  proof (induct n, simp)
	    case (Suc n)
	    show ?case
	    proof (cases "n \<ge> length ids")
	      case True
	      with Suc show ?thesis by simp
	    next
	      case False 
	      hence n: "n < length ids" by auto
	      let ?bef = "?ts (take n ids)"
	      let ?aft = "?ss (drop (Suc n) ids)"
	      let ?ssn = "pi (ss ! (ids ! n))"
	      let ?tsn = "pi (ts ! (ids ! n))"
	      from n ids have ids_n: "ids ! n < ?l" by force
	      have gn: "?both n = ?bef @ ?ssn # ?aft"
		by (simp only: drop_Suc_conv_tl[OF n, symmetric], simp, simp add: nth_map[of "ids ! n" ?bsa pi, simplified, OF ids_n] ss)
	      have gsn: "?both (Suc n) = ?bef @ ?tsn # ?aft"
		by (simp only: take_Suc_conv_app_nth[OF n], simp, simp add: nth_map[of "ids ! n" ?bta pi, simplified, OF ids_n] ts)
	      {
		assume "?ssn \<noteq> ?tsn \<or> ids ! n = i"
		with geq i have "(?ssn, ?tsn) \<in> S" by blast		
		from ctxt_closed_one[OF monoS this] 
		have gnsn: "(?gboth n, ?gboth (Suc n)) \<in> S"
		  by (simp only: gn gsn)
	      } note noteq_i = this
	      {
		assume "ids ! n \<noteq> i"
		{
		  assume "ids ! n < ?lb"
		  hence "?gboth (Suc n) = ?gboth n" by (simp add: gn gsn, simp add: ss ts, simp add: nth_append)
		}
		moreover
		{
		  assume i2: "ids ! n > ?lb"
		  have "?gboth (Suc n) = ?gboth n" unfolding term.simps gn gsn 
                    using i2 ss ts by (simp add: nth_append)
		}
		moreover
		{
		  from `ids ! n \<noteq> i` have "ids ! n < ?lb \<or> ids ! n > ?lb" by (simp only: idef, arith)
		}
		ultimately have "?gboth (Suc n) = ?gboth n" by blast
	      } note eq = this	      
	      show ?thesis 
	      proof (cases "i \<in> set (take (Suc n) ids)")
		case False 
		hence i_not: "i \<notin> set (take n ids)" and i_not2: "ids ! n \<noteq> i" using take_Suc_conv_app_nth[OF n] by auto
		from i_not Suc have rec: "?gss = ?gboth n" by blast 
		with False eq[OF i_not2] rec show ?thesis by simp 
	      next
		case True note oTrue = this
		show ?thesis
		proof (cases "i \<in> set (take n ids)")
		  case False
		  with True have id: "ids ! n = i" using take_Suc_conv_app_nth[OF n] by auto
		  from False Suc have rec: "?gss = ?gboth n" by blast
		  from noteq_i id have "(?gboth n, ?gboth (Suc n)) \<in> S" by blast
		  with True rec show ?thesis by simp
		next
		  case True
		  with Suc have rec: "(?gss, ?gboth n) \<in> S" by blast
		  show ?thesis
		  proof (cases "ids ! n = i")
		    case False
		    from rec eq[OF False] oTrue show ?thesis by simp
		  next
		    case True
		    with noteq_i have "(?gboth n, ?gboth (Suc n)) \<in> S" by blast
		    from rec this have "(?gss, ?gboth (Suc n)) \<in> S" using trans_S unfolding trans_def by blast
		    with oTrue show ?thesis by simp
		  qed
		qed
	      qed
	    qed
	  qed
	qed
	with spec[OF this, of "length ids"] iids have "(Fun g (?ss ids), Fun g (?ts ids)) \<in> S" by auto
	with g ss ts pi show "(Fun ?f (map (op ! (map ?pi bef @ ?pi s # map ?pi aft)) ids),
	  Fun ?f (map (op ! (map ?pi bef @ ?pi t # map ?pi aft)) ids)) \<in> S"
	  by auto 
      qed
    qed
  } note main = this
  show "ctxt_closed ?S"
    unfolding afs_rel_def afs_rule_def
    by (rule one_imp_ctxt_closed, insert main, auto)
qed
end

sublocale afs_redtriple \<subseteq> redtriple_order "afs_rel S" "afs_rel NS" "afs_rel NST"
proof -
  let ?S = "afs_rel S"
  let ?NS = "afs_rel NS"
  let ?NST = "afs_rel NST"
  show "redtriple_order ?S ?NS ?NST"
  proof
    show "refl ?NS"
      by (rule afs_refl[OF refl_NS])
  next
    show "refl ?NST"
      by (rule afs_refl[OF refl_NST])
  next
    show "trans ?S"
      by (rule afs_trans[OF trans_S])
  next
    show "trans ?NS"
      by (rule afs_trans[OF trans_NS])
  next
    show "trans ?NST"
      by (rule afs_trans[OF trans_NST])
  qed 
qed

sublocale afs_redtriple_impl \<subseteq> redtriple_impl "afs_rel S" "afs_rel NS" "afs_rel NST" "afs_check rS \<pi> cS" "afs_check rNS \<pi> cNS" "afs_check rNST \<pi> cNST"
proof -
  let ?S = "afs_rel S"
  let ?NS = "afs_rel NS"
  let ?NST = "afs_rel NST"
  let ?cS = "afs_check rS \<pi> cS"
  let ?cNS = "afs_check rNS \<pi> cNS"
  let ?cNST = "afs_check rNST \<pi> cNST"
  show "redtriple_impl ?S ?NS ?NST ?cS ?cNS ?cNST"
  proof
    fix lr
    assume "isOK (afs_check rS \<pi> cS lr)"
    thus "lr \<in> ?S" unfolding afs_check_def afs_rel_def 
      using checkSSound by (auto simp: split_paired_all Let_def afs_rule_def)
  next
    fix lr
    assume "isOK (afs_check rNS \<pi> cNS lr)"
    thus "lr \<in> ?NS" unfolding afs_check_def afs_rel_def 
      using checkNSSound by (auto simp: split_paired_all Let_def afs_rule_def)
  next
    fix lr
    assume "isOK (afs_check rNST \<pi> cNST lr)"
    thus "lr \<in> ?NST" unfolding afs_check_def afs_rel_def 
      using checkNSTSound by (auto simp: split_paired_all Let_def afs_rule_def)
  qed
qed


sublocale mono_afs_redtriple_impl \<subseteq> mono_redtriple_impl "afs_rel S" "afs_rel NS" "afs_rel NST" "afs_check rS \<pi> cS" "afs_check rNS \<pi> cNS" "afs_check rNST \<pi> cNST" "cm \<guillemotright> check (mono_afs \<pi>) mess"
proof
  assume "isOK (cm \<guillemotright> check (mono_afs \<pi>) mess)"
  with checkMonoSound have monoS: "ctxt_closed S" and mono_pi: "mono_afs \<pi>" by auto
  from ctxt_closed[OF mono_pi monoS] show "ctxt_closed (afs_rel S)" .
qed

definition permutation_afs :: "'f afs \<Rightarrow> bool" where
  "permutation_afs \<pi> = (\<forall> f n. \<exists> xs. \<pi> (f,n) = AFList xs \<and> set xs = {0 ..< n} \<and> distinct xs)"

context
  fixes \<pi> :: "'f afs"
  assumes perm: "permutation_afs \<pi>"
begin

lemma perm: "\<exists>xs. \<pi> (f, n) = AFList xs \<and> set xs = {0..<n} \<and> distinct xs"
  using perm[unfolded permutation_afs_def] by auto

lemma afs: "afs \<pi>"
proof (unfold_locales, unfold wf_afs_def, intro allI)
  fix f n
  show "wf_af_entry n (\<pi> (f, n))" using perm[of f n] by (cases "\<pi> (f,n)", auto)
qed

fun af_ctxt :: "('f,'v)ctxt \<Rightarrow> ('f,'v)ctxt" where
  "af_ctxt Hole = Hole"
| "af_ctxt (More f bef C aft) = (let i = length bef;
    n = Suc (i + length aft);
    ts = map (af_term \<pi>) (bef @ C\<langle>undefined\<rangle> # aft) in case \<pi> (f,n) of
      AFList xs \<Rightarrow> let i' = SOME i'. i' < n \<and> xs ! i' = i
      in More f (map (\<lambda> j. ts ! (xs ! j)) [0 ..< i']) (af_ctxt C) (map (\<lambda> j. ts ! (xs ! j)) [Suc i' ..< n]))"

lemma af_ctxt[simp]: "af_term \<pi> (C \<langle> t \<rangle>) = (af_ctxt C) \<langle> af_term \<pi> t \<rangle>"
proof (induct C)
  case (More f bef C aft)
  let ?i = "length bef"
  let ?n = "Suc (?i + length aft)"
  from perm[of f ?n] obtain xs where pi: "\<pi> (f,?n) = AFList xs" and xs: "set xs = {0 ..< ?n}" and d: "distinct xs" by auto
  from distinct_card[OF d] have len: "length xs = ?n" unfolding xs by simp
  let ?p = "\<lambda> i'. i' < ?n \<and> xs ! i' = ?i"
  def i' \<equiv> "SOME i'. ?p i'"
  have "?i \<in> set xs" unfolding xs by auto
  with len have "\<exists> i'. ?p i'" unfolding set_conv_nth by auto
  from someI_ex[OF this] have i': "?p i'" unfolding i'_def by simp
  let ?ts = "map (af_term \<pi>) (bef @ C\<langle>undefined\<rangle> # aft)"
  have "(af_ctxt (More f bef C aft)) \<langle> af_term \<pi> t \<rangle> = (More f (map (\<lambda>j. ?ts ! (xs ! j)) [0..<i']) (af_ctxt C)
      (map (\<lambda>j. ?ts ! (xs ! j)) [Suc i'..<?n]))\<langle>af_term \<pi> t\<rangle>" (is "?l = _")
    unfolding af_ctxt.simps Let_def pi af_entry.simps i'_def by simp
  also have "\<dots> = Fun f ((map (\<lambda>j. ?ts ! (xs ! j)) [0..<i']) @ af_term \<pi> (C \<langle> t \<rangle>) # (map (\<lambda>j. ?ts ! (xs ! j)) [Suc i'..<length xs]))"
    (is "_ = Fun f ?r") using More len by simp
  finally have id1: "?l = Fun f ?r" .  
  let ?is = "[0..<i'] @ i' # [Suc i' ..< length xs]" 
  have "xs = map (\<lambda> i. xs ! i) [0..<?n]" unfolding len[symmetric] by (rule nth_equalityI, auto)
  also have "[0..<?n] = ?is" using i' 
    by (metis len upt_append upt_conv_Cons)
  finally have id3: "(map (\<lambda> i. xs ! i) ?is) = xs" by simp
  have "af_term \<pi> (More f bef C aft)\<langle>t\<rangle> = 
    Fun f (map (\<lambda>x. map_term filtered_fun ((map (afs_term \<pi>) bef @ afs_term \<pi> C\<langle>t\<rangle> # map (afs_term \<pi>) aft) ! x)) xs)" 
    (is "?l = _") unfolding af_term_def by (simp add: pi o_def) 
  also have "\<dots> = Fun f (map (\<lambda>x. ((map (af_term \<pi>) bef @ af_term \<pi> C\<langle>t\<rangle> # map (af_term \<pi>) aft) ! x)) xs)"
    by (simp add: xs af_term_def[abs_def], auto simp: nth_append, case_tac "x - length bef", auto)
  also have "\<dots> = Fun f (map (\<lambda>x. ((map (af_term \<pi>) bef @ af_term \<pi> C\<langle>t\<rangle> # map (af_term \<pi>) aft) ! x)) (map (\<lambda> i. xs ! i) ?is))"
    (is "_ = Fun f ?r'") unfolding id3 ..
  finally have id2: "?l = Fun f ?r'" by simp
  have id4: "(Fun f ?r' = Fun f ?r) = (?r' = ?r)" by simp
  show ?case unfolding id1 id2 id4
  proof -
    note i' = i'[folded len]
    have i: "?i < length xs" unfolding len by simp
    show "?r' = ?r" 
    proof (rule nth_equalityI, simp, intro allI impI)
      fix j
      assume "j < length ?r'"
      from this[unfolded id3] have j: "j < length xs" by simp
      with xs have xsj: "xs ! j < length xs" unfolding len[symmetric] by auto
      show "?r' ! j = ?r ! j"
      proof (cases "j = i'")
        case True
        show ?thesis using i' unfolding True
          by (simp add: nth_append)
      next
        case False
        let ?ts' = "map (af_term \<pi>) (bef @ C\<langle>undefined\<rangle> # aft)"
        from False d j i' have xsji: "xs ! j \<noteq> ?i" 
          by (metis nth_eq_iff_index_eq)
        have "?r' ! j = ?ts' ! (xs ! j)" unfolding id3 nth_map[OF j]
          using xsji xsj unfolding len by (auto simp: nth_append)
        also have "\<dots> = ?r ! j" using False i' j by (simp add: nth_append)
        finally show ?thesis .
      qed
    qed
  qed
qed simp

lemma step_imp_afs_rstep: "(s,t) \<in> rstep R \<Longrightarrow> (af_term \<pi> s, af_term \<pi> t) \<in> rstep (af_rule \<pi> ` R)"
proof (induct rule: rstep_induct)
  case (IH C \<sigma> l r)
  interpret afs \<pi> by (rule afs)
  let ?a = "af_term \<pi>"
  let ?R = "af_rule \<pi> ` R"
  def s \<equiv> "?a (l \<cdot> \<sigma>)"
  def t \<equiv> "?a (r \<cdot> \<sigma>)"
  from IH have rule: "(?a l, ?a r) \<in> ?R" unfolding af_rule_def af_term_def afs_rule_def by force
  have "(?a (l \<cdot> \<sigma>), ?a (r \<cdot> \<sigma>)) \<in> rstep ?R"
    by (rule rstepI[OF rule, of _ Hole], auto simp: af_term_def)
  from rstep.ctxt[OF this, of "af_ctxt C"] show ?case by simp
qed

lemma steps_imp_afs_rsteps: "(s,t) \<in> (rstep R)^* \<Longrightarrow> (af_term \<pi> s, af_term \<pi> t) \<in> (rstep (af_rule \<pi> ` R))^*"
  using afs.af_rsteps[OF afs] .

lemma step_imp_afs_rel_rstep: assumes st: "(s,t) \<in> relto (rstep R) (rstep S)"
  shows "(af_term \<pi> s, af_term \<pi> t) \<in> relto (rstep (af_rule \<pi> ` R)) (rstep (af_rule \<pi> ` S))"
proof -
  let ?a = "af_term \<pi>"  
  from st obtain u v where su: "(s,u) \<in> (rstep S)^*" and uv: "(u,v) \<in> (rstep R)" and vt: "(v,t) \<in> (rstep S)^*" by auto
  from steps_imp_afs_rsteps[OF su] step_imp_afs_rstep[OF uv] steps_imp_afs_rsteps[OF vt]
  show ?thesis by blast
qed

lemma af_SN_relto_rstep: assumes SN: "SN (relto (rstep (af_rule \<pi> ` R)) (rstep (af_rule \<pi> ` S)))" (is "SN ?R'")
  shows "SN (relto (rstep R) (rstep S))" (is "SN ?R")
proof
  fix f
  assume steps: "\<forall> i. (f i, f (Suc i)) \<in> ?R"
  def g \<equiv> "\<lambda> i. (af_term \<pi> (f i))"
  from step_imp_afs_rel_rstep[OF spec[OF steps]]
  have "\<And> i. (g i, g (Suc i)) \<in> ?R'" unfolding g_def by auto
  with SN show False unfolding SN_defs by auto
qed

lemma af_SN_rstep: assumes SN: "SN (rstep (af_rule \<pi> ` R))"
  shows "SN (rstep R)" using af_SN_relto_rstep[of "{}" R] SN by simp
end
  

type_synonym 'f afs_list = "(('f \<times> nat) \<times> af_entry)list"

definition default_af_entry :: "nat \<Rightarrow> af_entry"
  where "default_af_entry n \<equiv> AFList [0..<n]"

definition afs_list_to_afs :: "('f :: key) afs_list \<Rightarrow> 'f afs" where
  "afs_list_to_afs \<pi> \<equiv> fun_of_map_fun (ceta_map_of \<pi>) (\<lambda> fn. default_af_entry (snd fn))"

definition afs_to_unused_arity :: "'f afs_list \<Rightarrow> nat" where
  "afs_to_unused_arity \<pi> \<equiv> foldr (\<lambda>((f, n), e) m. max n m) \<pi> 0"

lemma afs_to_unused_arity_sound: shows "\<forall> e \<in> set pi. snd (fst e) \<le> afs_to_unused_arity pi"
proof (induct pi)
  case (Cons e es)
  {
    fix f n en
    assume "((f,n),en) \<in> set (e # es)"
    hence "((f,n),en) = e \<or> ((f,n),en) \<in> set es" (is "?one \<or> ?two")
      by simp
    hence "n \<le> afs_to_unused_arity (e # es)"
    proof
      assume ?one thus ?thesis unfolding afs_to_unused_arity_def by auto
    next
      assume ?two with Cons have "n \<le> afs_to_unused_arity es" by auto
      also have "\<dots> \<le> afs_to_unused_arity (e # es)" unfolding afs_to_unused_arity_def by (induct es arbitrary: e, auto)
      finally show ?thesis .
    qed
  }
  thus ?case by force 
qed simp

lemma default_af_entry_id: 
  assumes "\<pi> (f,length ts) = default_af_entry (length ts)"
  shows "afs_term \<pi> (Fun f ts) = Fun (FPair f (length ts)) (map (afs_term \<pi>) ts)"
  by (simp add: default_af_entry_def assms Let_def, simp only: list_eq_iff_nth_eq, simp)

lemma wf_default_af_entry: 
  "wf_af_entry n (default_af_entry n)"
  unfolding default_af_entry_def by auto

definition check_wf_afs :: "('f::show) afs_list \<Rightarrow> shows check" where
  "check_wf_afs pi = check_allm (\<lambda>((f, n), e).
    (case e of
      Collapse i \<Rightarrow> check (i < n) i
    | AFList ids \<Rightarrow> check_all (\<lambda> i. i < n) ids)
    <+? (\<lambda>i. ''index '' +#+ shows i +@+ '' too high for AF-entry of'' +#+ shows(f,n) +@+ shows_nl))
    pi"

lemma check_wf_afs_sound:
  assumes check: "isOK (check_wf_afs pi)"
  shows "afs (afs_list_to_afs pi)"
proof (intro afs.intro, unfold wf_afs_def afs_list_to_afs_def Let_def ceta_map_of fun_of_map_fun.simps snd_conv, intro allI impI)
  fix f n
  show "wf_af_entry n (case map_of pi (f, n) of None \<Rightarrow> default_af_entry n | Some en \<Rightarrow> en)"
  proof (cases "map_of pi (f, n)", simp add: wf_default_af_entry)
    case (Some en)
    hence mem: "((f,n),en) \<in> set pi" by (rule map_of_SomeD)
    show ?thesis 
    proof (simp add: Some, cases en)
      case (Collapse i)
      with Some mem check show "wf_af_entry n en" unfolding check_wf_afs_def by auto
    next
      case (AFList ids)
      with Some mem check show "wf_af_entry n en" unfolding check_wf_afs_def by auto
    qed
  qed
qed

definition check_permutation_afs :: "('f::show) afs_list \<Rightarrow> shows check" where
  "check_permutation_afs pi = check_allm (\<lambda>((f, n), e).
    (case e of
      Collapse i \<Rightarrow> error (shows ''collapsing entry not allowed'')
    | AFList ids \<Rightarrow> check (distinct ids \<and> set ids = set [0 ..< n]) (shows ''every position needs to occur exactly once''))      
    <+? (\<lambda>s. shows ''argument filter is not a permutation for '' +@+ shows(f,n) +@+ shows_nl +@+ s))
    pi"

lemma check_permutation_afs: assumes ok: "isOK(check_permutation_afs pi)"
  shows "permutation_afs (afs_list_to_afs pi)"
unfolding permutation_afs_def
proof (intro allI)
  fix f n  
  let ?pi = "afs_list_to_afs pi"
  show "\<exists>xs. ?pi (f, n) = AFList xs \<and> set xs = {0..<n} \<and> distinct xs"
  proof (cases "map_of pi (f,n)")
    case None
    hence "?pi (f,n) = default_af_entry n" unfolding afs_list_to_afs_def
      by simp
    thus ?thesis unfolding default_af_entry_def by simp
  next
    case (Some entry)
    have "((f,n),entry) \<in> set pi" using map_of_SomeD[OF Some] .
    from ok[unfolded check_permutation_afs_def, simplified, rule_format, OF this]
    show ?thesis using Some unfolding afs_list_to_afs_def by (cases entry, auto)
  qed
qed

fun check_mono_afs_help where "check_mono_afs_help ((f,a),e) = (''entry for '' +#+ shows f +@+ '' not monotonic'' +#+ shows_nl)"

definition
  check_mono_afs :: "('f::show)afs_list \<Rightarrow> shows check"
where
  "check_mono_afs pi = (check_all (\<lambda>x . mono_af_entry (snd(fst x)) (snd x)) pi <+? check_mono_afs_help)"

lemma check_mono_afs_sound: assumes check: "isOK(check_mono_afs pi)"
  and check2: "wf_afs (afs_list_to_afs pi)"
  shows "mono_afs (afs_list_to_afs pi)"
proof (unfold mono_afs_def afs_list_to_afs_def Let_def ceta_map_of fun_of_map_fun.simps snd_conv, intro allI impI)
  fix f n
  show "mono_af_entry n (case map_of pi (f, n) of None \<Rightarrow> default_af_entry n | Some en \<Rightarrow> en)"
  proof (cases "map_of pi (f, n)")
    case None thus ?thesis by (simp add: wf_default_af_entry default_af_entry_def)
  next
    case (Some en)
    hence mem: "((f,n),en) \<in> set pi" by (rule map_of_SomeD)
    show ?thesis 
    proof (simp add: Some, cases en)
      case (Collapse i)
      with Some mem check check2 show "mono_af_entry n en" unfolding check_mono_afs_def mono_af_entry_def wf_afs_def by auto
    next
      case (AFList ids)
      with Some mem check check2 show "mono_af_entry n en" unfolding check_mono_afs_def mono_af_entry_def wf_afs_def by auto
    qed
  qed
qed

definition afs_list_to_af :: "('f :: key) afs_list \<Rightarrow> 'f af" where
  "afs_list_to_af pi \<equiv> fun_of_map_fun' (ceta_map_of pi) (\<lambda> _ _. True) (\<lambda> e.
      (case e of
        Collapse j \<Rightarrow> (op = j)
      | AFList ids \<Rightarrow> (\<lambda>i. i \<in> set ids)))"

context afs
begin

lemma apply_af_entry_const_is_id: "apply_af_entry (FPair f (0 :: nat)) (\<pi> (f, 0)) [] = Fun (FPair f 0) []"
proof -
  let ?e = "\<pi> (f, 0)"
  from wf_afs have "wf_af_entry 0 ?e" unfolding wf_afs_def by auto
  thus ?thesis by (cases ?e, simp_all)
qed
end

lemma af_compatible:
  fixes pi :: "('f :: key) afs_list"
  assumes wf_afs: "wf_afs (afs_list_to_afs pi)" (is "wf_afs ?pi")
  and refl_NS: "refl NS"
  and all_ctxt_closed: "all_ctxt_closed UNIV NS"
  shows "af_compatible (afs_list_to_af pi) (afs.afs_rel ?pi (NS :: ('f filtered,'v)term rel))" (is "af_compatible ?af ?NS")
  unfolding af_compatible_def
proof (intro allI)      
  fix f :: 'f and bef :: "('f,'v)term list" and s t :: "('f,'v)term" and  aft
  let ?pi = "afs_list_to_afs pi"
  let ?bsa = "bef @ s # aft"
  let ?bta = "bef @ t # aft"
  let ?lb = "length bef"
  let ?l = "Suc (?lb + length aft)"    
  let ?s = "Fun f ?bsa"
  let ?t = "Fun f ?bta"
  show "?af (f, ?l) ?lb \<or> (?s,?t) \<in> ?NS"
  proof (cases "?af (f, ?l) ?lb")
    case False
    let ?e = "?pi (f, ?l)"
    from wf_afs have wf: "wf_af_entry ?l ?e" unfolding wf_afs_def by auto
    from wf_afs have afs: "afs ?pi" by (rule afs.intro) 
    show ?thesis
    proof (cases "map_of pi (f, ?l)")
      case None
      with False show ?thesis unfolding afs_list_to_af_def by auto
    next
      case (Some e)
      hence e: "?e = e" by (simp add: afs_list_to_afs_def)
      let ?ls = "length (bef @ s # aft)"
      let ?lt = "length (bef @ t # aft)"
      let ?msa = "map (afs_term ?pi) ?bsa"
      let ?mta = "map (afs_term ?pi) ?bta"
      let ?pfs = "apply_af_entry (FPair f ?l) ?e ?msa"
      let ?pft = "apply_af_entry (FPair f ?l) ?e ?mta"
      have all2: "\<forall> i < ?l. i \<noteq> ?lb \<longrightarrow> (?msa ! i, ?mta ! i) \<in> NS" 
      proof (intro allI impI)
	fix i
	assume "i < ?l" and ib: "i \<noteq> ?lb"
	hence i_s: "i < ?ls" and i_t: "i < ?lt" by simp_all
	{
	  assume bef: "i < length bef"
	  with refl_NS have "(?msa ! i, ?mta ! i) \<in> NS" unfolding refl_on_def 
	    by (simp only: nth_map[OF i_s], simp add: nth_append)
	}
	moreover
	{
	  assume i: "i > length bef"
	  from this obtain k where k1: "i - length bef = Suc k" by (cases "i - length bef", arith+)
	  with i_s have k2: "k < length aft" by auto
	  from i refl_NS have "(?msa ! i, ?mta ! i) \<in> NS" unfolding refl_on_def 
	    by (simp only: nth_map[OF i_s] nth_map[OF i_t], simp add: nth_append k1)
	} 
	moreover
	{
	  have "i < length bef \<or> i = length bef \<or> i > length bef" by auto
	}      
	ultimately
	show "(?msa ! i, ?mta ! i) \<in> NS" using ib by blast
      qed 
      have "(?pfs,?pft) \<in> NS" 
      proof (cases ?e)
	case (Collapse i)
	with wf have len: "i < ?l"  by simp
	from Collapse[simplified] Some False have "i \<noteq> ?lb" unfolding afs_list_to_afs_def afs_list_to_af_def by auto
	with all2 len Collapse show ?thesis by auto
      next
	case (AFList ids)
	with wf have ids: "\<forall> i \<in> set ids. i < ?l" by simp
	show ?thesis 
	proof (simp only: AFList apply_af_entry.simps)
	  let ?msi = "map (op ! (map (afs_term ?pi) ?bsa)) ids"
	  let ?mti = "map (op ! (map (afs_term ?pi) ?bta)) ids"
	  from ids False AFList[simplified] Some have "\<forall> i \<in> set ids. i \<noteq> ?lb" unfolding afs_list_to_afs_def afs_list_to_af_def 
	    by (auto)
	  with ids all2 have args: "\<forall> i < length ids. (?msi ! i, ?mti ! i) \<in> NS" by auto
	  show "(Fun (FPair f ?l) ?msi, Fun (FPair f ?l) ?mti) \<in> NS"
	    by (rule all_ctxt_closedE[OF all_ctxt_closed], insert args, auto)
	qed
      qed
      thus "afs_list_to_af pi (f, ?l) ?lb \<or> (Fun f ?bsa, Fun f ?bta) \<in> afs.afs_rel ?pi NS" unfolding afs.afs_rel_def[OF afs] afs_rule_def by (simp add: afs)
    qed
  qed simp
qed

lemma af_ce_compat:
  fixes pi :: "('f :: {show,key}) afs_list"
    and NS :: "('f filtered, 'v :: show) term rel"
  assumes wf_afs: "wf_afs (afs_list_to_afs pi)" 
    and ce_compat: "\<exists> n. \<forall> m \<ge> n. \<forall> c d. ce_trs (c,d,m) \<subseteq> NS"
  shows "\<exists> n. \<forall> m \<ge> n. \<forall> c d. ce_trs (c,d,m) \<subseteq> afs.afs_rel (afs_list_to_afs pi) NS"
proof -
  let ?pi = "afs_list_to_afs pi"
  let ?NS = "afs.afs_rel ?pi NS"
  let ?cS = "afs_check rS ?pi cS"
  let ?cNS = "afs_check rNS ?pi cNS"
  let ?ar = "afs_to_unused_arity pi"
  let ?sar = "Suc (Suc ?ar)"
  let ?af = "afs_list_to_af pi"
  have pi2: "\<And> m c. m \<ge> ?sar \<Longrightarrow> ?pi (c, m) = default_af_entry m"
  proof -
    fix m c
    assume m: "m \<ge> ?sar"
    show "?pi (c,m) = default_af_entry m"
    proof (unfold afs_list_to_afs_def Let_def ceta_map_of fun_of_map_fun.simps snd_conv, cases "map_of pi (c, m)")
      case (Some foo)
      hence "((c,m),foo) \<in> set pi" by (rule map_of_SomeD)
      from bspec[OF afs_to_unused_arity_sound this] m
      show "(case map_of pi (c, m) of None \<Rightarrow> default_af_entry m | Some foo \<Rightarrow> foo) = (default_af_entry m)" by simp
    qed simp
  qed
  from wf_afs have afs: "afs ?pi" by (rule afs.intro) 
  from ce_compat obtain n where ce_NS: "\<And> m c d . m \<ge> n \<Longrightarrow> ce_trs (c,d,m) \<subseteq> NS" by blast
  let ?n = "max n ?sar"
  show "\<exists> n. \<forall> m \<ge> n. \<forall> c d. ce_trs (c,d,m) \<subseteq> ?NS"
  proof (rule exI[of _ ?n], intro allI impI)
    fix m c d
    assume m: "m \<ge> ?n"
    hence mn: "m \<ge> n" by simp
    {
      fix t s :: "('f,'v)term"
      let ?list = "t # s # replicate m (Fun d [])"
      from pi2 m have id: "?pi (c,length ?list) = default_af_entry (length ?list)" by simp
      have "(afs_term ?pi (Fun c ?list), afs_term ?pi t) \<in> NS"
      	using ce_NS[OF mn, simplified ce_trs.simps, of "FPair c (Suc (Suc m))" "FPair d 0"]  
        by (simp only: default_af_entry_id[of ?pi c ?list, OF id], auto simp: afs.apply_af_entry_const_is_id[OF afs])
    }
    moreover
    {
      fix t s :: "('f,'v)term"
      let ?list = "t # s # replicate m (Fun d [])"
      from pi2 m have id: "?pi (c,length ?list) = default_af_entry (length ?list)" by simp
      have "(afs_term ?pi (Fun c ?list), afs_term ?pi s) \<in> NS"
      using ce_NS[OF mn, simplified ce_trs.simps, of "FPair c (Suc (Suc m))" "FPair d 0"]  
        by (simp only: default_af_entry_id[of ?pi c ?list, OF id], auto simp: afs.apply_af_entry_const_is_id[OF afs])
    }
    ultimately
    show "ce_trs (c,d,m) \<subseteq> ?NS"
      by (auto simp: ce_trs.simps afs.afs_rel_def[OF afs] afs_rule_def pi2)      
  qed
qed



locale afs_ce_redtriple_impl = afs_redtriple_impl "afs_list_to_afs pi" S NS NST cS cNS cNST + filtered: redtriple_impl S NS NST cS cNS cNST + ce: ce_redtriple S NS NST + order: redtriple_order S NS NST
  for pi and S NS NST :: "(('f::{show,key})filtered,'v::show)trs" and cS cNS cNST

locale afs_ce_mono_redtriple_impl = mono_afs_redtriple_impl "afs_list_to_afs pi" S NS NST cS cNS cNST cm + filtered: ce_mono_redtriple_impl S NS NST cS cNS cNST cm + order: redtriple_order S NS NST
  for pi and S NS NST :: "(('f::{show,key})filtered,'v::show)trs" and cS cNS cNST cm

sublocale afs_ce_mono_redtriple_impl \<subseteq> afs_ce_redtriple_impl ..

sublocale afs_ce_redtriple_impl \<subseteq> ce_af_redtriple_impl "afs_rel S" "afs_rel NS" "afs_rel NST" "afs_check rS (afs_list_to_afs pi) cS" "afs_check rNS (afs_list_to_afs pi) cNS" "afs_check rNST (afs_list_to_afs pi) cNST" "afs_list_to_af pi"
proof -
  let ?pi = "afs_list_to_afs pi"
  let ?S = "afs_rel S"
  let ?NS = "afs_rel NS"
  let ?NST = "afs_rel NST"
  let ?cS = "afs_check rS ?pi cS"
  let ?cNS = "afs_check rNS ?pi cNS"
  let ?cNST = "afs_check rNST ?pi cNST"
  let ?af = "afs_list_to_af pi"
  show "ce_af_redtriple_impl ?S ?NS ?NST ?cS ?cNS ?cNST ?af"
  proof
    have rp1: "pre_order_pair S NS" ..
    have rp2: "redpair_order S NS" ..
    show "af_compatible ?af ?NS"
      by (rule af_compatible[OF wf_afs pre_order_pair.refl_NS[OF rp1] redpair_order.all_ctxt_closed[OF rp2]])
  next
    show "\<exists> n. \<forall> m \<ge> n. \<forall> c d. ce_trs (c,d,m) \<subseteq> ?NS"
      by (rule af_ce_compat[OF wf_afs NS_ce_compat])
  qed
qed


sublocale afs_ce_mono_redtriple_impl \<subseteq> ce_mono_redtriple_impl "afs_rel S" "afs_rel NS" "afs_rel NST" "afs_check rS (afs_list_to_afs pi) cS" "afs_check rNS (afs_list_to_afs pi) cNS" "afs_check rNST (afs_list_to_afs pi) cNST" "cm \<guillemotright> check_mono_afs pi"
proof 
  assume "isOK (cm \<guillemotright> check_mono_afs pi)"
  hence c1: "isOK(cm)" and c2: "isOK(check_mono_afs pi)" by auto
  hence check: "isOK(cm \<guillemotright> check (mono_afs (afs_list_to_afs pi)) mess)"
    using check_mono_afs_sound[OF c2 wf_afs] by auto 
  show "ctxt_closed (afs.afs_rel (afs_list_to_afs pi) S)"
    by (rule checkMonoSound[OF check])
next
  let ?S = "afs_rel S"
  show "\<exists> n. \<forall> m \<ge> n. \<forall> c d. ce_trs (c,d,m) \<subseteq> ?S"
    by (rule af_ce_compat[OF wf_afs S_ce_compat])
qed


fun shows_afs :: "('f::show)afs_list \<Rightarrow> shows"
where "shows_afs af = shows_map (\<lambda>((f,n),e). ''pi('' +#+ shows f +@+ ''/'' +#+ shows n +@+ '') = '' +#+ (case e of
  Collapse i \<Rightarrow> shows(Suc i)
| AFList ids \<Rightarrow> shows(map Suc ids)) +@+ shows_nl) af"


definition af_redtriple :: "'f afs_list \<Rightarrow> ('f filtered,'v)redtriple \<Rightarrow> ('f :: {show,key},'v :: show)redtriple"
where "af_redtriple pi rp \<equiv> let af = afs_list_to_af pi; afs = afs_list_to_afs pi in \<lparr>
  redtriple.valid = do {check_wf_afs pi; redtriple.valid rp}, 
  s = afs_check (shows_string ''>'') afs (redtriple.s rp),
  ns = afs_check (shows_string ''>='') afs (redtriple.ns rp),
  nst = afs_check (shows_string ''>='') afs (redtriple.nst rp),
  af = af,
  mono = \<lambda> cs. do {check_mono_afs pi; redtriple.mono rp (map (afs_rule afs) cs)} ,
  desc = (shows_string ''Argument Filter: '') +@+ shows_nl +@+ shows_afs pi +@+ shows_nl +@+ redtriple.desc rp,
  cpx = no_complexity
  \<rparr>"

lemma af_redtriple: 
  fixes pi :: "('f :: {show,key}) afs_list" and
        rp :: "'a \<Rightarrow> ('f filtered, 'v :: show)redtriple"
  assumes rp: "generic_redtriple_impl rp" shows 
  "generic_redtriple_impl (\<lambda> I. (af_redtriple pi (rp I)))"
proof
  fix I s_list ns_list nst_list
  let ?rp = "af_redtriple pi (rp I) :: ('f, 'v)redtriple"
  assume valid: "isOK(redtriple.valid ?rp)" and stri: "isOK (check_allm (redtriple.s ?rp) s_list)" and nstri: "isOK (check_allm (redtriple.ns ?rp) ns_list)"
    and nstrit: "isOK (check_allm (redtriple.nst ?rp) nst_list)"
  let ?pi = "afs_list_to_afs pi"
  let ?p = "afs_list_to_af pi"
  let ?pi' = "redtriple.af (rp I)"
  let ?pi'' = "redtriple.af (af_redtriple pi (rp I))"
  let ?cpx = "redtriple.cpx ?rp"
  have cpx: "?cpx = no_complexity" unfolding af_redtriple_def Let_def by simp
  have pi'': "?pi'' = ?p"  unfolding af_redtriple_def Let_def by simp
  let ?ar = "afs_to_unused_arity pi"
  let ?sar = "Suc (Suc ?ar)"
  let ?af = "afs_list_to_af pi"
  let ?af_term = "afs_term ?pi"
  let ?af_list = "map (\<lambda> (s,t). (?af_term s, ?af_term t))"
  let ?as_list = "?af_list s_list"
  let ?ans_list = "?af_list ns_list"
  let ?anst_list = "?af_list nst_list"
  from valid have valid: "isOK(redtriple.valid (rp I))" and wf_afs: "isOK(check_wf_afs pi)" unfolding af_redtriple_def by (auto simp: Let_def)
  from wf_afs interpret afs ?pi by (rule check_wf_afs_sound)
  from stri have stri: "isOK(check_allm (redtriple.s (rp I)) ?as_list)" by (auto simp: Let_def af_redtriple_def afs_check_def)
  from nstri have nstri: "isOK(check_allm (redtriple.ns (rp I)) ?ans_list)" by (auto simp: Let_def af_redtriple_def afs_check_def)
  from nstrit have nstrit: "isOK(check_allm (redtriple.nst (rp I)) ?anst_list)" by (auto simp: Let_def af_redtriple_def afs_check_def)
  from generic_redtriple_impl.generate_redtriple[of rp, OF rp valid stri nstri nstrit]
  obtain S NS NST where redtriple: "cpx_ce_af_redtriple_order S NS NST ?pi' (redtriple.cpx (rp I))"
     and stri: "set ?as_list \<subseteq> S"
     and nstri: "set ?ans_list \<subseteq> NS"
     and nstrit: "set ?anst_list \<subseteq> NST"
     and cmono: "isOK (redtriple.mono (rp I) (?as_list @ ?ans_list @ ?anst_list)) \<longrightarrow> mono_ce_af_redtriple_order S NS NST ?pi' \<and> ctxt_closed S" by auto
  let ?aS = "afs_rel S"
  let ?aNS = "afs_rel NS"
  let ?aNST = "afs_rel NST"
  from redtriple interpret cpx_ce_af_redtriple_order S NS NST ?pi' "redtriple.cpx (rp I)" .
  have rp1: "pre_order_pair S NS" ..
  have rp2: "redpair_order S NS" ..
  interpret afs_redtriple ?pi S NS NST ..
  interpret cpx_ce_af_redtriple_order ?aS ?aNS ?aNST ?p ?cpx unfolding cpx
  proof(unfold_locales, rule af_ce_compat[OF wf_afs NS_ce_compat])
    show "af_compatible ?p ?aNS"
      by (rule af_compatible[OF wf_afs pre_order_pair.refl_NS[OF rp1] redpair_order.all_ctxt_closed[OF rp2]])
  qed
  show "\<exists>S NS NST. cpx_ce_af_redtriple_order S NS NST (redtriple.af (af_redtriple pi (rp I))) ?cpx \<and>
                 set s_list \<subseteq> S \<and>
                 set ns_list \<subseteq> NS \<and>
                 set nst_list \<subseteq> NST \<and>
                 (isOK (redtriple.mono (af_redtriple pi (rp I)) (s_list @ ns_list @ nst_list)) \<longrightarrow>
                  mono_ce_af_redtriple_order S NS NST ?pi'' \<and> ctxt_closed S)" unfolding pi''
  proof (intro exI conjI)
    show "set s_list \<subseteq> ?aS" using stri unfolding afs_rel_def afs_rule_def by auto
  next
    show "set ns_list \<subseteq> ?aNS" using nstri unfolding afs_rel_def afs_rule_def by auto
  next
    show "set nst_list \<subseteq> ?aNST" using nstrit unfolding afs_rel_def afs_rule_def by auto
  next
    show "cpx_ce_af_redtriple_order ?aS ?aNS ?aNST ?p ?cpx" ..
  next
    show "isOK (redtriple.mono (af_redtriple pi (rp I)) (s_list @ ns_list @ nst_list)) \<longrightarrow>
                  mono_ce_af_redtriple_order ?aS ?aNS ?aNST ?p \<and> ctxt_closed ?aS" (is "?prem \<longrightarrow> ?mono \<and> ?ctxt")
    proof
      assume prem: ?prem
      have afs: "afs_rule ?pi = (\<lambda> (l,r). (afs_term ?pi l, afs_term ?pi r))"
        by (rule ext, unfold afs_rule_def, force)
      from prem cmono have mono: "mono_ce_af_redtriple_order S NS NST ?pi'" and cmono: "isOK(check_mono_afs pi)" and ctxt: "ctxt_closed S" unfolding af_redtriple_def by (auto simp: Let_def afs) 
      from mono interpret mono_ce_af_redtriple_order S NS NST ?pi' .
      have ?mono 
        by (unfold_locales, rule af_ce_compat[OF wf_afs S_ce_compat])
      have ?ctxt
        by (rule ctxt_closed[OF check_mono_afs_sound[OF cmono wf_afs] ctxt])
      from `?mono` `?ctxt` show "?mono \<and> ?ctxt" ..
    qed
  qed
qed

end
