(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Conditional_Rewriting_Impl
imports
  Conditional_Rewriting
  QDP_Framework_Impl
begin

type_synonym ('f, 'v) crules = "('f, 'v) crule list"

definition create_ctxts :: "('f, 'v) rules \<Rightarrow> (nat \<Rightarrow> ('f, 'v) ctxt) option"
where
  "create_ctxts R =
    (case R of
      [] \<Rightarrow> None
    | Cons l RR \<Rightarrow> do {
      cs \<leftarrow> Option_Monad.mapM (\<lambda>(l, r).
        (case l of
          Fun U (t # ts) \<Rightarrow> Some (More U [] \<box> ts)
        | _ \<Rightarrow> None)) RR;
      let n = length cs;
      Some (\<lambda>i. if i < length cs then cs ! i else \<box>) })"

definition
  create_U ::
    "(('f, 'v) crule \<times> ('f, 'v) rules) list \<Rightarrow> (('f, 'v) crule \<Rightarrow> nat \<Rightarrow> ('f, 'v) ctxt) option"
where
  "create_U c_rs = do {
    cr_ctxts \<leftarrow> Option_Monad.mapM (\<lambda> (cr, rs). do {
      guard (length rs = Suc (length (snd cr)));
      ctxt \<leftarrow> create_ctxts rs;
      Some (cr, ctxt)
    }) c_rs;
    let m = map_of cr_ctxts;
    Some (\<lambda> cr. case m cr of None \<Rightarrow> (\<lambda> _. \<box>) | Some ctxt \<Rightarrow> ctxt)
  }"

fun shows_conditions :: "('f :: show, 'v :: show) rule list \<Rightarrow> bool \<Rightarrow> shows"
where
  "shows_conditions [] _ = id" |
  "shows_conditions (x # xs) b = (shows (if b then '' | '' else '', '') +@+ 
    shows_rule' shows shows '' ->* '' x +@+ shows_conditions xs False)"

definition shows_crule :: "('f :: show, 'v :: show) crule \<Rightarrow> shows"
where
  "shows_crule cr = (shows_rule (fst cr) +@+ shows_conditions (snd cr) True)"

definition
  rules_impl :: "(('f, 'v) crule \<Rightarrow> nat \<Rightarrow> ('f, 'v) ctxt) \<Rightarrow> ('f, 'v) crule \<Rightarrow> ('f, 'v) rules"
where
  "rules_impl U cr = map (\<lambda>i.
    (unraveling.lhs_n U cr i, unraveling.rhs_n U cr i)) [0 ..< Suc (length (snd cr))]"

declare unraveling.lhs_n.simps[code] unraveling.rhs_n.simps[code]

lemma rules_impl[simp]: "set (rules_impl U cr) = unraveling.rules U cr"
proof -
  interpret unraveling "{}" U .
  obtain n where n: "Suc (length (snd cr)) = n" by auto
  have "{0..<Suc (length (snd cr))} = {i. i < Suc (length (snd cr))}" (is "?l = _")
    unfolding n by auto
  also have "... = {i. i \<le> length (snd cr)}" (is "_ = ?r") by auto
  finally have id: "?l = ?r" .
  show ?thesis unfolding rules_def rules_impl_def 
    by (simp del: upt_Suc add: id)
qed

definition check_unraveling :: "(('f :: show,'v :: show)crule \<times> ('f,'v)rules)list \<Rightarrow> ('f,'v)crules \<Rightarrow> ('f,'v)rules result"
  where "check_unraveling c_rs ctrs \<equiv> do {
      check_subseteq ctrs (map fst c_rs) <+? 
        (\<lambda> cr. shows ''did not find rule '' +@+ shows_crule cr +@+ shows_nl);
      U \<leftarrow> case create_U c_rs of None \<Rightarrow> error (shows ''unable to extract unraveling contexts'') | Some U \<Rightarrow> return U;
      check_allm (\<lambda> (c,rs). check (rules_impl U c = rs) (shows ''problem with rules of '' +@+ shows_crule c +@+ shows_nl)) c_rs;
      return (concat (map snd c_rs))
   } <+? (\<lambda> e. shows ''problem in unraveling '' +@+ shows_nl +@+ e)"

lemma check_unraveling: assumes ok: "check_unraveling c_rs ctrs = return r"
  and SN: "SN (rstep (set r))"
  shows "quasi_reductive (set ctrs)"
proof - 
  note ok = ok[unfolded check_unraveling_def, simplified]
  from ok obtain U where U: "create_U c_rs = Some U" by (cases "create_U c_rs", auto)
  note ok = ok[unfolded U, simplified]
  from ok have r: "r = concat (map snd c_rs)" by simp
  interpret unraveling "set ctrs" U .
  show ?thesis
  proof (rule SN_UR_quasi_reductive, rule SN_subset[OF SN rstep_mono])
    show "UR \<subseteq> set r" 
    proof(rule subsetI)
      fix lr
      assume lr: "lr \<in> UR"
      from this[unfolded UR_def] obtain cr where cr: "cr \<in> set ctrs"
      and lr: "lr \<in> rules cr" by auto
      from ok cr obtain rs where mem: "(cr,rs) \<in> set c_rs" by auto
      from ok mem have "rules_impl U cr = rs" by auto
      from arg_cong[OF this, of set] have "rules cr = set rs" by simp
      with lr have lr: "lr \<in> set rs" by auto
      with mem show "lr \<in> set r" unfolding r by force
    qed
  qed
qed      

hide_const (open) rules_impl

end

