(*
Copyright 2009-2013 René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Countables
imports
  Term
  Labelings
  "$AFP/Datatype_Order_Generator/Countable_Generator"
begin

datatype ('a,'b)large_count = 
  Count_1 'a 
| Count_2 'b "('a,'b)large_count list"
| Count_3 "('a,'b)large_count" 'b
| Count_4 "('a,'b)large_count" "('a,'b)large_count list"
| Count_5 "('a,'b)large_count"

fun term_large_count where
  "term_large_count (Var x) = Count_1 x"
| "term_large_count (Fun f ts) = Count_2 f (map term_large_count ts)"

fun lab_large_count where
  "lab_large_count (Lab ll l) = Count_3 (lab_large_count ll) l"
| "lab_large_count (FunLab ll lls) = Count_4 (lab_large_count ll) (map lab_large_count lls)"
| "lab_large_count (UnLab b) = Count_1 b"
| "lab_large_count (Sharp ll) = Count_5 (lab_large_count ll)"

derive countable large_count
instantiation "term" :: (countable,countable)countable
begin
instance
proof
  from ex_inj obtain f :: "('b,'a)large_count \<Rightarrow> nat" where inj: "inj f" ..
  have "inj (term_large_count :: ('a,'b)term \<Rightarrow> ('b,'a)large_count)" unfolding inj_on_def
  proof (intro ballI impI)
    fix s t
    assume "term_large_count s = term_large_count t"
    thus "s = t"
    proof (induct s arbitrary: t)
      case (Var x t)
      thus ?case by (cases t, auto)
    next
      case (Fun f ss t)
      then obtain ts where t: "t = Fun f ts" by (cases t, auto)
      from Fun(2)[unfolded t, simplified] Fun(1)
      show ?case unfolding t by (induct ss arbitrary: ts, auto)
    qed
  qed
  from inj_comp[OF inj this] show "\<exists> g :: ('a,'b)term \<Rightarrow> nat. inj g" by blast
qed
end

instantiation "lab" :: (countable,countable)countable
begin
instance
proof
  from ex_inj obtain f :: "('a,'b)large_count \<Rightarrow> nat" where inj: "inj f" ..
  have "inj (lab_large_count :: ('a,'b)lab \<Rightarrow> ('a,'b)large_count)" unfolding inj_on_def
  proof (intro ballI impI)
    fix s t
    assume "lab_large_count s = lab_large_count t"
    thus "s = t"
    proof (induct s arbitrary: t)
      case (UnLab f)
      thus ?case by (cases t, auto)
    next
      case (Sharp s t)
      thus ?case by (cases t, auto)
    next
      case (Lab s l t)
      thus ?case by (cases t, auto)
    next
      case (FunLab f ss t)
      then obtain h ts where t: "t = FunLab h ts" by (cases t, auto)
      from FunLab(3)[unfolded t]
      have id1: "lab_large_count f = lab_large_count h" and id2: "map lab_large_count ss = map lab_large_count ts" by auto
      from FunLab(1)[OF id1] have f: "f = h" .
      from FunLab(2) id2 have "ss = ts" by (induct ss arbitrary: ts, auto)
      with f t show ?case by auto
    qed
  qed
  from inj_comp[OF inj this] show "\<exists> g :: ('a,'b)lab \<Rightarrow> nat. inj g" by blast
qed
end
 
end
