(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Innermost_Usable_Rules_Impl
imports
  Innermost_Usable_Rules 
  Icap_Impl
  QDP_Framework_Impl
  Term_Order_Impl
  Inductive_Set_Impl
begin

fun is_ur_closed_term_impl where
  "is_ur_closed_term_impl NFQ e_cap R U S (Var x) = True"
| "is_ur_closed_term_impl NFQ e_cap R U S (Fun f ts) = 
    ((\<forall> t \<in> set ts. is_ur_closed_term_impl NFQ e_cap R U S t) \<and>                
    (\<forall>(l, r) \<in> set R. (l, r) \<in> set U \<or>
      (case mgu_class (Fun f (map e_cap ts)) l of
        None \<Rightarrow> True
      | Some \<mu> \<Rightarrow> \<not> (\<forall> u \<in> set (args l). NFQ (mv_yvar u \<cdot> \<mu>)) \<or> \<not> (\<forall> u \<in> set S. NFQ (u \<cdot> \<mu>)))))"

fun is_ur_closed_term_af_impl 
  where "is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S (Var x) = True"
     |  "is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S (Fun f ts) = (let n = length ts; \<pi>f = \<pi> (f,n) in
              ((\<forall> (i,t) \<in> set (zip [0 ..< n] ts). \<pi>f i \<longrightarrow> is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S t) \<and>                
              (\<forall> (l,r) \<in> set R. (l,r) \<in> set U \<or> (case mgu_class (Fun f (map e_cap ts)) l of
                 None \<Rightarrow> True | Some \<mu> \<Rightarrow> \<not> (\<forall> u \<in> set (args l). NFQ (mv_yvar u \<cdot> \<mu>)) \<or> \<not> (\<forall> u \<in> set S. NFQ (u \<cdot> \<mu>))))))"

lemma is_ur_closed_term_impl: fixes R :: "('f,string)rules" and Q :: "('f,string)term list"
  assumes NFQ: "NFQ = (\<lambda> t. t \<in> NF_terms (set Q))"
  and e_cap: "e_cap = icap (set R) (set Q) (set S)"
  shows "is_ur_closed_term_impl NFQ e_cap R U S = is_ur_closed_term' (set R) (set U) (set Q) icap full_af (set S)"
proof -
  let ?is_ur_closed_term = "is_ur_closed_term' (set R) (set U) (set Q) icap full_af"
  {
    fix t
    have "is_ur_closed_term_impl NFQ e_cap R U S t = ?is_ur_closed_term (set S) t"
    proof (induct t)
      case (Fun f ts)
      hence ball: "Ball (set ts) (is_ur_closed_term_impl NFQ e_cap R U S) = 
         (\<forall>i<length ts.
        full_af (f, length ts) i \<longrightarrow> ?is_ur_closed_term (set S) (ts ! i))" unfolding full_af_def[of "(f,length ts)"] set_conv_nth[of ts] by auto
      have all: "\<And> p q. \<lbrakk> \<And> l r. p l r = q l r\<rbrakk> \<Longrightarrow> (\<forall> (l,r) \<in> set R. (l,r) \<in> set U \<or> p l r) = (\<forall> (l,r) \<in> set R. (l,r) \<in> set U \<or> q l r)" by auto
      have ecap: "icap (set R) (set Q) (set S) = e_cap" unfolding e_cap ..
      have conj: "\<And> x y z u. x = y \<Longrightarrow> z = u \<Longrightarrow> (x \<and> z) = (y \<and> u)" by auto
      {
        fix x f 
        have "(case x of None \<Rightarrow> True | Some y \<Rightarrow> f y) = (\<not> (\<exists> y. x = Some y \<and> \<not> f y))" by (cases x, auto)
      } note id = this
      show ?case 
        unfolding is_ur_closed_term_impl.simps is_ur_closed_term'.simps
        by (rule conj[OF ball all], unfold ecap NFQ id, simp) blast
    qed simp
  } note main = this
  show ?thesis
    by (rule ext, rule main)
qed

lemma is_ur_closed_term_af_impl: fixes R :: "('f,string)rules" and Q :: "('f,string)term list"
  assumes NFQ: "NFQ = (\<lambda> t. t \<in> NF_terms (set Q))"
  and e_cap: "e_cap = icap (set R) (set Q) (set S)"
  shows "is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S = is_ur_closed_term' (set R) (set U) (set Q) icap \<pi> (set S)"
proof -
  let ?is_ur_closed_term = "is_ur_closed_term' (set R) (set U) (set Q) icap \<pi>"
  {
    fix t
    have "is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S t = ?is_ur_closed_term (set S) t"
    proof (induct t)
      case (Fun f ts)
      hence ball: "(\<forall> i < length ts. \<pi> (f,length ts) i \<longrightarrow> is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S (ts ! i)) = 
         (\<forall>i<length ts.
        \<pi> (f, length ts) i \<longrightarrow> ?is_ur_closed_term (set S) (ts ! i))" (is "?l' = ?r")
        using Fun[unfolded set_conv_nth[of ts]] by auto
      have id: "?l' = (\<forall> (i,t) \<in> set (zip [0 ..< length ts] ts). \<pi> (f,length ts) i \<longrightarrow> is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S t)" unfolding set_zip by force
      note ball = ball[unfolded id]
      have all: "\<And> p q. \<lbrakk> \<And> l r. p l r = q l r\<rbrakk> \<Longrightarrow> (\<forall> (l,r) \<in> set R. (l,r) \<in> set U \<or> p l r) = (\<forall> (l,r) \<in> set R. (l,r) \<in> set U \<or> q l r)" by auto
      have ecap: "icap (set R) (set Q) (set S) = e_cap" unfolding e_cap ..
      have conj: "\<And> x y z u. x = y \<Longrightarrow> z = u \<Longrightarrow> (x \<and> z) = (y \<and> u)" by auto
      {
        fix x f 
        have "(case x of None \<Rightarrow> True | Some y \<Rightarrow> f y) = (\<not> (\<exists> y. x = Some y \<and> \<not> f y))" by (cases x, auto)
      } note id = this
      show ?case 
        unfolding is_ur_closed_term_af_impl.simps is_ur_closed_term'.simps Let_def
        by (rule conj[OF ball all], unfold ecap NFQ id, simp) blast
    qed simp
  } note main = this
  show ?thesis
    by (rule ext, rule main)
qed

definition is_ur_closed_term_impl_mv 
  where "is_ur_closed_term_impl_mv NFQ e_cap R U S \<equiv> let urc = is_ur_closed_term_impl NFQ e_cap R U (map mv_xvar S) in (\<lambda> t. urc (mv_xvar t))"

definition is_ur_closed_term_af_impl_mv 
  where "is_ur_closed_term_af_impl_mv NFQ e_cap \<pi> R U S \<equiv> let urc = is_ur_closed_term_af_impl NFQ e_cap \<pi> R U (map mv_xvar S) in (\<lambda> t. urc (mv_xvar t))"

lemma is_ur_closed_term_impl_mv: fixes R :: "('f,string)rules" and Q :: "('f,string)term list"
  assumes NFQ: "NFQ = (\<lambda> t. t \<in> NF_terms (set Q))"
  and e_cap: "e_cap = icap (set R) (set Q) (set (map mv_xvar S))"
  shows "is_ur_closed_term_impl_mv NFQ e_cap R U S t = is_ur_closed_term_mv' (set R) (set U) (set Q) icap' full_af (set S) t" 
  unfolding
    is_ur_closed_term_impl_mv_def 
    is_ur_closed_term_impl[OF NFQ e_cap]
    set_map
    is_ur_closed_term_mv_icap' Let_def ..

lemma is_ur_closed_term_af_impl_mv: fixes R :: "('f,string)rules" and Q :: "('f,string)term list"
  assumes NFQ: "NFQ = (\<lambda> t. t \<in> NF_terms (set Q))"
  and e_cap: "e_cap = icap (set R) (set Q) (set (map mv_xvar S))"
  shows "is_ur_closed_term_af_impl_mv NFQ e_cap \<pi> R U S t = is_ur_closed_term_mv' (set R) (set U) (set Q) icap' \<pi> (set S) t" 
  unfolding
    is_ur_closed_term_af_impl_mv_def 
    is_ur_closed_term_af_impl[OF NFQ e_cap]
    set_map
    is_ur_closed_term_mv_icap' Let_def ..

definition is_ur_closed_impl_dpp_mv   :: "('d, 'f :: {key,show}, string) dpp_ops \<Rightarrow> 'd
  \<Rightarrow> ('f,string)rules \<Rightarrow> ('f,string)term list \<Rightarrow> ('f,string)term \<Rightarrow> bool"
  where "is_ur_closed_impl_dpp_mv I d \<equiv>
  let ic = icap_impl_dpp I d;
      qnf = dpp_ops.is_QNF I d;
      r = dpp_ops.rules I d;
      urc = (\<lambda> S. is_ur_closed_term_impl qnf (ic S) r)
    in (\<lambda> U S. let S' = map mv_xvar S in (\<lambda> t. urc S' U S' (mv_xvar t)))"

definition is_ur_closed_af_impl_dpp_mv   :: "('d, 'f :: {key,show}, string) dpp_ops \<Rightarrow> 'd \<Rightarrow> 'f af
  \<Rightarrow> ('f,string)rules \<Rightarrow> ('f,string)term list \<Rightarrow> ('f,string)term \<Rightarrow> bool"
  where "is_ur_closed_af_impl_dpp_mv I d \<pi> \<equiv>
  let ic = icap_impl_dpp I d;
      qnf = dpp_ops.is_QNF I d;
      r = dpp_ops.rules I d;
      urc = (\<lambda> S. is_ur_closed_term_af_impl qnf (ic S) \<pi> r)
    in (\<lambda> U S. let S' = map mv_xvar S in (\<lambda> t. urc S' U S' (mv_xvar t)))"

lemma is_ur_closed_impl_dpp_mv:
  fixes I::"('d, 'f::{show,key}, string) dpp_ops" and d::"'d"
  defines q: "q \<equiv> dpp_ops.Q I d"
      and r: "r \<equiv> dpp_ops.rules I d"
  assumes I: "dpp_spec I"
  shows "is_ur_closed_impl_dpp_mv I d u s t = is_ur_closed_term_mv' (set r) (set u) (set q) icap' full_af (set s) t" 
proof -
  interpret dpp_spec I by fact
  show ?thesis
    unfolding is_ur_closed_term_impl_mv[OF refl refl, symmetric]
      is_ur_closed_impl_dpp_mv_def Let_def is_QNF_sound icap_impl_dpp_icap[OF I] r q 
    unfolding is_ur_closed_term_impl_mv_def Let_def ..
qed    

lemma is_ur_closed_af_impl_dpp_mv:
  fixes I::"('d, 'f::{show,key}, string) dpp_ops" and d::"'d"
  defines q: "q \<equiv> dpp_ops.Q I d"
      and r: "r \<equiv> dpp_ops.rules I d"
  assumes I: "dpp_spec I"
  shows "is_ur_closed_af_impl_dpp_mv I d \<pi> u s t = is_ur_closed_term_mv' (set r) (set u) (set q) icap' \<pi> (set s) t" 
proof -
  interpret dpp_spec I by fact
  show ?thesis
    unfolding is_ur_closed_term_af_impl_mv[OF refl refl, symmetric]
      is_ur_closed_af_impl_dpp_mv_def Let_def is_QNF_sound icap_impl_dpp_icap[OF I] r q 
    unfolding is_ur_closed_term_af_impl_mv_def Let_def ..
qed    

definition
  usable_rules_proc ::
    "('dpp, 'f :: {key,show}, string) dpp_ops \<Rightarrow> 
     ('f,string)rules \<Rightarrow> 'dpp proc"
where
  "usable_rules_proc I U dpp \<equiv> 
    check_return (do {
     check (dpp_ops.NFQ_subset_NF_rules I dpp) (shows ''innermost rewriting required'');
     check (dpp_ops.nfs I dpp \<or> dpp_ops.minimal I dpp \<or> dpp_ops.wwf_rules I dpp) (shows ''normal form subst, minimality or well-formedness required'');
     let P = dpp_ops.pairs I dpp;
     let urc = is_ur_closed_impl_dpp_mv I dpp U;
     let check_urc = (\<lambda> S t. check (urc S t) (shows ''term '' +@+ shows t +@+ shows '' is not closed under usable rules''));
     let nfs = dpp_ops.nfs I dpp;
     check_allm (\<lambda>(l, r). do {
        (if nfs then succeed else check_subseteq (vars_term_list r) (vars_term_list l) <+? (\<lambda> x. shows ''variable condition in P violated''));
        check_urc [l] r
     }) P;       
     check_allm (\<lambda>(l,r). check_urc (args l) r) U
   })
   (dpp_ops.intersect_rules I dpp U)"


lemma usable_rules_proc_main: assumes I: "dpp_spec I"
  and ok: "usable_rules_proc I U d = return d'"
  shows "R_Q_U_ecap (set (dpp_ops.rules I d)) (set (dpp_ops.Q I d)) icap'"
  and "R_Q_U_ecap.usable_rules_precond (set (dpp_ops.rules I d)) (set U) (set (dpp_ops.Q I d)) icap' (dpp_ops.dpp I d)"
proof -
  interpret dpp_spec I by fact
  let ?P = "set (P d)"
  let ?Pw = "set (Pw d)"
  let ?Q = "set (Q d)"
  let ?R = "set (R d)"
  let ?Rw = "set (Rw d)"
  let ?U = "set U"
  let ?Rb = "?R \<union> ?Rw"
  let ?Rb' = "set (rules d)" 
  note ok = ok[unfolded usable_rules_proc_def dpp_spec_sound Let_def
    is_ur_closed_impl_dpp_mv[OF I], simplified] 
  from ok have NF: "NF_terms ?Q \<subseteq> NF_trs ?Rb" and NF': "NF_terms ?Q \<subseteq> NF_trs ?Rb'" by auto
  show "R_Q_U_ecap ?Rb' ?Q icap'"
    by (unfold_locales, rule icap, rule NF')
  interpret R_Q_U_ecap ?Rb' ?U ?Q icap' by fact
  show "usable_rules_precond (dpp_ops.dpp I d)" 
    unfolding dpp_sound
    unfolding usable_rules_precond.simps
  proof (intro allI conjI impI)
    have Rb: "?Rb = ?Rb'" by simp
    show "?Rb \<subseteq> ?Rb'" by simp
  next
    fix l r
    assume lr: "(l,r) \<in> ?P \<union> ?Pw" 
    from lr ok show "is_ur_closed_term_mv full_af {l} r" by auto
    assume "\<not> NFS d"
    with lr ok 
    show "vars_term r \<subseteq> vars_term l" by force
  next
    fix l r
    assume "(l,r) \<in> ?U"
    with ok
    show "is_ur_closed_term_mv full_af (set (args l)) r" by auto    
  qed (insert ok, auto)
qed

  
lemma usable_rules_proc: assumes I: "dpp_spec I"
 shows "dpp_spec.sound_proc_impl I (usable_rules_proc I (U :: ('f :: {key,show},string)rules))"
proof -
  from assms interpret dpp_spec I .
  show ?thesis
  proof
    fix d d'    
    assume ok: "usable_rules_proc I U d = Inr d'" and fin: "finite_dpp (dpp d')"
    note us = usable_rules_proc_main[OF I ok]    
    let ?Q = "set (Q d)"
    let ?U = "set U"
    interpret R_Q_U_ecap "set (rules d)" ?U ?Q icap' by fact
    note ok = ok[unfolded usable_rules_proc_def dpp_spec_sound Let_def] 
    from ok have d': "d' = intersect_rules d U" by auto
    have d': "dpp d' = (NFS d, M d, set (P d), set (Pw d), ?Q, set (R d) \<inter> ?U, set (Rw d) \<inter> ?U)" 
      unfolding d' unfolding intersect_rules_sound ..
    note fin = fin[unfolded this]
    show "finite_dpp (dpp d)" unfolding dpp_spec_sound
      by (rule usable_rules_proc[OF fin us(2)[unfolded dpp_sound]])
  qed
qed 

definition
  mono_inn_usable_rules_ce_proc ::
    "('dpp, 'f, string) dpp_ops \<Rightarrow> ('f::{key,show}, string) redtriple \<Rightarrow> 
     ('f,string)rules \<Rightarrow> ('f,string)rules \<Rightarrow> ('f,string)rules \<Rightarrow> 'dpp proc"
where
  "mono_inn_usable_rules_ce_proc I rp Premove Rremove ur dpp = check_return (do {
     usable_rules_proc I ur dpp;
     let P = dpp_ops.pairs I dpp;
     let us = \<Union> set (map (funas_term o snd) (P @ ur));
     let filt = (\<lambda> lr. (\<forall> f \<in> funas_term (fst lr). f \<in> us));
     let (pms, pns) = dpp_ops.split_pairs I dpp Premove;
     let (ps, pnwf) = partition filt pms;
     let (urms, urns) = partition (\<lambda> u. u \<in> set Rremove) ur;
     let (urs, urnwf) = partition filt urms;
     redtriple.valid rp;
     redtriple.mono rp (ps @ urs @ urns @ urnwf @ pns @ pnwf);
     check_allm (redtriple.ns rp) (urns @ urnwf)
       <+? (\<lambda>s. ''problem when orienting usable rules'' +#+ shows_nl +@+ s);
     check_allm (redtriple.s rp) urs
       <+? (\<lambda>s. ''problem when orienting usable rules'' +#+ shows_nl +@+ s);
     check_allm (redtriple.ns rp) (pns @ pnwf)
       <+? (\<lambda>s. ''problem when orienting DPs'' +#+ shows_nl +@+ s);
     check_allm (redtriple.s rp) ps
       <+? (\<lambda>s. ''problem when orienting DPs'' +#+ shows_nl +@+ s)
   } <+? (\<lambda>s. ''could not apply the innermost usable rules reduction pair processor with the following'' +#+ shows_nl +@+
     (redtriple.desc rp) +@+ shows_nl +@+ s))
   (dpp_spec.delete_pairs_rules I dpp Premove Rremove)"

lemma mono_inn_usable_rules_ce_proc: assumes I: "dpp_spec I"
  and grp: "generic_redtriple_impl grp"
  shows "dpp_spec.sound_proc_impl I (mono_inn_usable_rules_ce_proc I (grp rp) ps rs ur)"
proof -
  interpret dpp_spec I by fact
  show ?thesis
  proof
  fix d d'
  assume fin: "finite_dpp (dpp d')"
    and ok: "mono_inn_usable_rules_ce_proc I (grp rp) ps rs ur d = return d'"
    let ?rp = "grp rp"
    let ?S = "set ps"
    let ?P = "set (P d)"
    let ?Pw = "set (Pw d)"
    let ?S' = "(?P \<union> ?Pw) \<inter> ?S"
    let ?Pb = "set (pairs d)"
    let ?Q = "set (Q d)"
    let ?R = "set (R d)"
    let ?Rw = "set (Rw d)"
    let ?Sr = "set rs"
    let ?U = "set ur"
    interpret generic_redtriple_impl grp by (rule grp)
    obtain us where us: "us = \<Union> ( set (map (funas_term o snd) (dpp_ops.pairs I d @ ur)))" by auto
    let ?filt = "\<lambda> lr. (\<forall> f \<in> funas_term (fst lr). f \<in> us)"
    let ?WF = "{lr. ?filt lr}"
    obtain Pms Pns where p1: "dpp_ops.split_pairs I d ps = (Pms,Pns)" by force
    obtain Ps Pnwf where p2: "partition ?filt Pms = (Ps,Pnwf)" by force
    obtain Rms Rns where r1: "partition (\<lambda> r. r \<in> ?Sr) ur = (Rms,Rns)" by force
    obtain Rs Rnwf where r2: "partition ?filt Rms = (Rs,Rnwf)" by force    
    from r1 r2 have ur: "?U = set Rs \<union> set Rnwf \<union> set Rns" and Rs: "set Rs = (?U \<inter> ?Sr) \<inter> ?WF" by auto
    from split_pairs_sound[OF p1] have Pms: "set Pms = (?P \<union> ?Pw) \<inter> ?S" and Pns: "set Pns = (?P \<union> ?Pw) - ?S" by auto
    from p2 have Ps: "set Ps = ((?P \<union> ?Pw) \<inter> ?S) \<inter> ?WF" and Pnwf: "set Pnwf = ((?P \<union> ?Pw) \<inter> ?S) - ?WF" unfolding Pms[symmetric] by auto
    note P = Ps Pnwf Pns
    note ok = ok[unfolded mono_inn_usable_rules_ce_proc_def Let_def p1 r1 split p2[unfolded us] r2[unfolded us], simplified]
    from ok obtain dpp' where urp: "usable_rules_proc I ur d = return dpp'" by (cases "usable_rules_proc I ur d", auto)
    note urp = usable_rules_proc_main[OF I urp]
    interpret R_Q_U_ecap "set (rules d)" ?U ?Q icap' by fact
    let ?all = "(Ps @ Rs) @ (Rns @ Rnwf @ Pns @ Pnwf) @ []"
    from ok have valid: "isOK(redtriple.valid ?rp)" 
        and mono: "isOK(redtriple.mono ?rp ?all)"
        and NS: "isOK (check_allm (redtriple.ns ?rp) (Rns @ Rnwf @ Pns @ Pnwf))"
        and S: "isOK(check_allm (redtriple.s ?rp) (Ps @ Rs))" 
        and d': "d' = dpp_spec.delete_pairs_rules I d ps rs"
          by auto
    let ?us = "(\<Union> ( (funas_term o snd) ` (?U \<union> ?Pb)))"
    from generate_redtriple[OF valid S NS, of Nil] mono
    obtain S NS NST where "mono_ce_af_redtriple_order S NS NST (redtriple.af ?rp)" and 
      S: "set Ps \<union> set Rs \<subseteq> S" and NS: "set Rns \<union> set Rnwf \<union> set Pns \<union> set Pnwf \<subseteq> NS" 
      and mono: "ctxt_closed S"
      by auto
    then interpret old: mono_ce_af_redtriple_order S NS NST "redtriple.af ?rp" by simp
    have redp: "mono_ce_af_redtriple S NS NST full_af"
      by (unfold_locales, rule full_af)
    from S NS have ur: "?U \<subseteq> NS \<union> S" unfolding ur by auto
    have P: "?P \<union> ?Pw = set Ps \<union> set Pns \<union> set Pnwf" unfolding P by blast
    from S NS have p: "?P \<union> ?Pw \<subseteq> NS \<union> S" unfolding P by auto
    let ?NWF = "{lr | lr. \<not> funas_term (fst lr) \<subseteq> ?us}"
    have swap: "?U \<union> ?Pb = ?Pb \<union> ?U" by auto
    have "?S' - ?NWF \<subseteq> set Ps" unfolding Ps us pairs_sound[symmetric] swap by auto
    also have "... \<subseteq> S" using S by auto
    finally have Ps: "?S' - ?NWF \<subseteq> S" .
    have "(?Sr \<inter> ?U) - ?NWF = (?U \<inter> ?Sr) - ?NWF" by auto
    also have "... \<subseteq> set Rs" unfolding Rs us swap by auto
    also have "... \<subseteq> S" using S by auto
    finally have Rs: "(?Sr \<inter> ?U) - ?NWF \<subseteq> S" .
    from p ur have or: "?P \<union> ?Pw \<union> ?U \<subseteq> NS \<union> S" by auto
    show "finite_dpp (dpp d)" unfolding dpp_sound
      by (rule usable_rules_finite_dpp_ce[OF urp(2)[unfolded dpp_sound] 
       redp mono or _ _ refl fin[unfolded d' delete_simps]], insert Rs Ps, auto)
  qed
qed


fun ur_term :: "('f,string)trs \<Rightarrow> ('f,string)terms \<Rightarrow> ('f,string)cap_fun \<Rightarrow> 'f af \<Rightarrow> ('f,string)terms \<Rightarrow> ('f,string)term \<Rightarrow> ('f,string)trs"
  where "ur_term R Q ecap \<pi> S (Var x) = {}"
     |  "ur_term R Q ecap \<pi> S (Fun f ts) = (let rec = map (ur_term R Q ecap \<pi> S) ts in 
             \<Union>{rec ! i | i. i < length ts \<and> \<pi> (f,length ts) i} \<union> { (l,r) | l r \<mu>. (l,r) \<in> R \<and> mgu_class (Fun f (map (ecap R Q S) ts)) l = Some \<mu> \<and> (S \<union> mv_yvar ` (set (args l))) \<cdot>set \<mu> \<subseteq> NF_terms Q })"

lemma ur_term_subset: "ur_term R Q ecap \<pi> S t \<subseteq> R"
proof (induct t)
  case (Fun f ts)
  {
    fix i
    assume "i < length ts"
    hence "ts ! i \<in> set ts" by auto
    from Fun[OF this] have "ur_term R Q ecap \<pi> S (ts ! i) \<subseteq> R" .
  }
  thus ?case by force
qed auto

lemma is_ur_closed'_ur_term: assumes U: "ur_term R Q ecap \<pi> S t \<subseteq> U"
  shows "is_ur_closed_term' R U Q ecap \<pi> S t"
  using U 
proof (induct t)
  case (Var x) thus ?case by auto
next
  case (Fun f ts)
  let ?P = "\<lambda> l \<mu>. mgu_class (Fun f (map (ecap R Q S) ts)) l = Some \<mu> \<and>
    S \<union> mv_yvar ` (set (args l)) \<cdot>set \<mu> \<subseteq> NF_terms Q"
  {
    fix l r
    assume "(l,r) \<in> R"
    with Fun(2)[simplified]
    have "(l,r) \<in> U \<or> \<not> (\<exists> \<mu>. ?P l \<mu>)" by blast
  } note two = this
  show ?case unfolding is_ur_closed_term'.simps
  proof (intro conjI ballI allI impI)
    fix i
    assume i: "i < length ts"
    and \<pi>: "\<pi> (f,length ts) i"
    from i have mem: "ts ! i \<in> set ts" by auto
    from Fun(2)[simplified] i \<pi> have sub: "map (ur_term R Q ecap \<pi> S) ts ! i \<subseteq> U"
      by blast
    show "is_ur_closed_term' R U Q ecap \<pi> S (ts ! i)"
      by (rule Fun(1)[OF mem], insert sub i, simp)
  qed (insert two, auto)
qed

lemma ur_term_is_ur_closed': assumes U: "is_ur_closed_term' R U Q ecap \<pi> S t"
  shows "ur_term R Q ecap \<pi> S t \<subseteq> U"
  using U
proof (induct t)
  case (Var x) thus ?case by auto
next
  case (Fun f ts)
  let ?P = "\<lambda> l \<mu>. mgu_class (Fun f (map (ecap R Q S) ts)) l = Some \<mu> \<and>
    S \<union> mv_yvar ` (set (args l)) \<cdot>set \<mu> \<subseteq> NF_terms Q"
  {
    fix l r
    assume "(l,r) \<in> \<Union> {map (ur_term R Q ecap \<pi> S) ts ! i | i. i < length ts \<and> \<pi> (f,length ts) i}"
    then obtain i where i: "i < length ts" and \<pi>: "\<pi> (f,length ts) i" and lr: "(l,r) \<in> map (ur_term R Q ecap \<pi> S) ts ! i" by auto
    from lr i have lr: "(l,r) \<in> ur_term R Q ecap \<pi> S (ts ! i)" by auto
    also have "... \<subseteq> U"
      by (rule Fun(1), insert \<pi> i Fun(2), auto)
    finally have "(l,r) \<in> U" .
  } note one = this
  let ?P = "\<lambda> l \<mu>. mgu_class (Fun f (map (ecap R Q S) ts)) l = Some \<mu> \<and>
    S \<union> mv_yvar ` (set (args l)) \<cdot>set \<mu> \<subseteq> NF_terms Q"
  {
    fix l r
    assume "(l,r) \<in> { u. \<exists> l r \<mu>. u = (l,r) \<and> (l,r) \<in> R \<and> ?P l \<mu>}"
    then obtain \<mu> where lr: "(l,r) \<in> R" and P: "?P l \<mu>" by auto
    with Fun(2)
    have "(l,r) \<in> U" by auto
  } note two = this
  show ?case using one two by auto
qed

lemma ur_term_least_is_ur_closed': "ur_term R Q ecap \<pi> S t = 
  (LEAST U. is_ur_closed_term' R U Q ecap \<pi> S t)"
  by (rule Least_equality[symmetric],
    rule is_ur_closed'_ur_term[OF subset_refl],
    rule ur_term_is_ur_closed')

fun ur_term_impl 
  where "ur_term_impl NFQ e_cap R \<pi> S (Var x) = []"
     |  "ur_term_impl NFQ e_cap R \<pi> S (Fun f ts) = (let n = length ts; rec = map (ur_term_impl NFQ e_cap R \<pi> S) ts in 
             remdups (concat (map (\<lambda> (i,urs). if \<pi> (f,n) i then urs else []) (zip [0 ..< n] rec)) @ filter (\<lambda> (l,r). 
  case mgu_class (Fun f (map e_cap ts)) l of
                 None \<Rightarrow> False | Some \<mu> \<Rightarrow> \<forall> u \<in> set (S @ map mv_yvar (args l)). NFQ (u \<cdot> \<mu>)
   ) R))"

lemma ur_term_impl: fixes R Q ecap
  assumes NFQ: "NFQ = (\<lambda> t. t \<in> NF_terms Q)"
  and e_cap: "e_cap = ecap (set R) Q (set S)"
  shows "set (ur_term_impl NFQ e_cap R \<pi> S t) = ur_term (set R) Q ecap \<pi> (set S) t"
proof (induct t)
  case (Var x)
  show ?case by simp
next
  case (Fun f ts)
  let ?impl = "ur_term_impl NFQ e_cap R \<pi> S"
  let ?sem = "ur_term (set R) Q ecap \<pi> (set S)"
  let ?base = "set (filter (\<lambda> (l,r). 
  case mgu_class (Fun f (map e_cap ts)) l of
                 None \<Rightarrow> False | Some \<mu> \<Rightarrow> \<forall> u \<in> set (S @ map mv_yvar (args l)). NFQ (u \<cdot> \<mu>)) R)"
  let ?base2 = " {(l,r). (l,r) \<in> set R \<and> (\<exists> \<mu>.
        mgu_class (Fun f (map (ecap (set R) Q (set S)) ts)) l = Some \<mu> \<and>
        set S \<union> mv_yvar ` set (args l) \<cdot>set \<mu> \<subseteq> NF_terms Q)}"
  have "set (?impl (Fun f ts)) = 
          {x. (\<exists> i. i < length ts \<and> \<pi> (f, length ts) i \<and> x \<in> set (?impl (ts ! i)))}
         \<union> ?base" (is "_ = ?rec \<union> _")
    by (auto simp: Let_def set_zip)
  also have "?rec = {x. (\<exists> i. i < length ts \<and> \<pi> (f, length ts) i \<and> x \<in> ?sem (ts ! i))}" 
    (is "_ = ?rec2")
    using Fun[unfolded set_conv_nth[of ts]] by blast
  also have "?base = ?base2" unfolding NFQ e_cap[symmetric]
    by (force split: option.splits)
  finally have id: "set (?impl (Fun f ts)) = ?rec2 \<union> ?base2" .
  show "set (?impl (Fun f ts)) = ?sem (Fun f ts)"
    unfolding id ur_term.simps Let_def by auto
qed

context
  fixes R :: "('f,string)trs"
  and Q :: "('f,string)terms"
  and ecap :: "('f,string)cap_fun"
  and \<pi> :: "'f af"
begin
abbreviation (input) calc where "calc \<equiv> (\<lambda> (argsl,t) ur. ur \<in> ur_term R Q ecap \<pi> (mv_xvar ` set argsl) (mv_xvar t))"
abbreviation (input) nxt where "nxt \<equiv> (\<lambda> (l,r). {(args l, r)})"

definition "usable_rules_pair_mv \<equiv> \<lambda> (s,t). {ur . generic_inductive_set.the_set R 
    calc nxt ([s],t) ur}"

lemma usable_rules_pair_mv_ur_closed_t: 
   "is_ur_closed_term' R (usable_rules_pair_mv (s,t)) Q ecap \<pi> (mv_xvar ` {s}) (mv_xvar t)"
  by (rule is_ur_closed'_ur_term, unfold usable_rules_pair_mv_def split, 
  auto intro!: generic_inductive_set.non_rec, rule set_mp[OF ur_term_subset])

lemma usable_rules_pair_mv_ur_closed_U: assumes lr: "(l,r) \<in> usable_rules_pair_mv (s,t)"
   shows "is_ur_closed_term_mv' R (usable_rules_pair_mv (s,t)) Q ecap \<pi> (set (args l)) r"
proof -
  interpret generic_inductive_set R calc nxt .
  note d = usable_rules_pair_mv_def split
  from lr[unfolded d]
  have "the_set ([s],t) (l,r)" by simp
  note rec = rec_rec[OF this]
  show ?thesis
  proof (rule is_ur_closed'_ur_term, rule)
    fix l' r'
    assume ur: "(l',r') \<in> ur_term R Q ecap \<pi> (mv_xvar ` set (args l)) (mv_xvar r)"
    with ur_term_subset have "(l',r') \<in> R" by force
    note non_rec = non_rec[OF this]
    show "(l',r') \<in> usable_rules_pair_mv (s, t)"
      unfolding d
      by (simp, rule rec, simp, rule non_rec, insert ur, auto)
  qed
qed

end

hide_const calc
hide_const nxt

definition usable_rules_calculator where
  "usable_rules_calculator NFQ e_cap R \<pi> \<equiv> 
    let
      urt = (\<lambda> (S,t). let S' = map mv_xvar S; U = ur_term_impl NFQ (e_cap S') R \<pi> S' (mv_xvar t) in 
        (\<lambda> u. u \<in> set U));
      urules = map (\<lambda> (l,r). (args l,r)) R;
      ufun = precompute_fun urt urules
    in (\<lambda>(s, t). inductive_set_impl R ufun (\<lambda> (l,r). [(args l,r)]) [([s],t)])"

lemma usable_rules_calculator:
  fixes R Q ecap
  assumes NFQ: "NFQ = (\<lambda> t. t \<in> NF_terms Q)"
    and e_cap: "\<And> S. e_cap S = ecap (set R) Q (set S)"
  shows "set (usable_rules_calculator NFQ e_cap R \<pi> (s,t)) = usable_rules_pair_mv (set R) Q ecap \<pi> (s,t)"
  unfolding usable_rules_calculator_def Let_def split usable_rules_pair_mv_def 
  using ur_term_impl[OF NFQ, of _ ecap, OF e_cap]
  by (intro inductive_set_impl_single, auto)

definition get_ur_calc_impl_dpp 
  where "get_ur_calc_impl_dpp I d \<pi> \<equiv>
  let ic = icap_impl_dpp I d;
      qnf = dpp_ops.is_QNF I d;
      r = dpp_ops.rules I d
    in usable_rules_calculator qnf ic r \<pi>"

lemma get_ur_calc_impl_dpp:
  fixes I::"('d, 'f::{show,key}, string) dpp_ops" and d::"'d" and \<pi> s t
  defines q: "q \<equiv> dpp_ops.Q I d"
      and R: "R \<equiv> dpp_ops.rules I d"
  assumes I: "dpp_spec I"
      and NF: "NF_terms (set q) \<subseteq> NF_trs (set R)"
      and U: "U = get_ur_calc_impl_dpp I d \<pi> (s,t)"
  shows "\<exists> ecap. R_Q_U_ecap (set R) (set q) ecap \<and>
           (\<forall> l r. (l,r) \<in> set U \<longrightarrow> is_ur_closed_term' (set R) (set U) (set q) ecap \<pi> (mv_xvar ` set (args l)) (mv_xvar r)) \<and>
           is_ur_closed_term' (set R) (set U) (set q) ecap \<pi> (mv_xvar ` {s}) (mv_xvar t)"proof -
  interpret dpp_spec I by fact
  let ?R = "set (rules d)"
  let ?Q = "set (Q d)"
  show ?thesis unfolding R q
  proof (intro exI, rule conjI)
    show "R_Q_U_ecap ?R ?Q icap'"
      by (unfold_locales, rule icap, insert NF R q, auto)
  next
    have icap: "icap_impl_dpp I d = (\<lambda> s. icap ?R ?Q (set s))" unfolding R q
      unfolding icap_impl_dpp_icap[OF I, symmetric] ..
    note U = U[unfolded get_ur_calc_impl_dpp_def Let_def, unfolded is_QNF_sound icap]
    hence U: "set U = set (usable_rules_calculator (\<lambda>t. t \<in> NF_terms ?Q)
      (\<lambda>s. icap ?R ?Q (set s)) (rules d) \<pi> (s, t))" by simp
    also have "... = usable_rules_pair_mv ?R ?Q icap \<pi> (s,t)"
      by (rule usable_rules_calculator, auto)
    finally have U: "set U = usable_rules_pair_mv ?R ?Q icap \<pi> (s,t)" .
    note sU = U[symmetric]
    note closed_U = usable_rules_pair_mv_ur_closed_U[of _ _ ?R ?Q icap \<pi> s t, unfolded sU]
    note closed_t = usable_rules_pair_mv_ur_closed_t[of ?R ?Q icap \<pi> s t, unfolded sU] 
    note id = is_ur_closed_term_af_impl_mv[OF refl refl, unfolded is_ur_closed_term_af_impl_mv_def Let_def]
    let ?one = "(\<forall> l r. (l,r) \<in> set U \<longrightarrow> is_ur_closed_term' ?R (set U) ?Q icap' \<pi> (mv_xvar ` set (args l)) (mv_xvar r))"
    let ?two = "is_ur_closed_term' ?R (set U) ?Q icap' \<pi> (mv_xvar ` {s}) (mv_xvar t)"
    from closed_t have two: ?two unfolding is_ur_closed_term_mv_icap' .
    from closed_U have one: ?one unfolding is_ur_closed_term_mv_icap' by auto
    from one two show "?one \<and> ?two" by simp
  qed
qed

lemma get_ur_calc_impl_dpp_steps:
  fixes I::"('d, 'f::{show,key}, string) dpp_ops" and d::"'d"
  defines q: "q \<equiv> dpp_ops.Q I d"
      and R: "R \<equiv> dpp_ops.rules I d"
      and nfs: "nfs \<equiv> dpp_ops.nfs I d"
      and m: "m \<equiv> dpp_ops.minimal I d"
  assumes I: "dpp_spec I"
      and U: "U = get_ur_calc_impl_dpp I d full_af (s,t)"
      and NF: "NF_terms (set q) \<subseteq> NF_trs (set R)"
      and NFs: "s \<cdot> \<sigma> \<in> NF_terms (set q)"
      and nfsigma: "NF_subst nfs (s,t) \<sigma> (set q)"
      and varsR: "nfs \<or> SN_on (qrstep nfs (set q) (set R)) {t \<cdot> \<sigma>} \<or> wwf_qtrs (set q) (set R)"
      and vars: "\<not> nfs \<Longrightarrow> vars_term t \<subseteq> vars_term s"
      and steps: "(t \<cdot> \<sigma> ,u) \<in> (qrstep nfs (set q) (set R))^*"
  shows "(t \<cdot> \<sigma>, u) \<in> (qrstep nfs (set q) (set U))^*"
proof -
  interpret dpp_spec I by fact
  let ?R = "set (rules d)"
  let ?Q = "set (Q d)"
  let ?nfs = "NFS d"
  from get_ur_calc_impl_dpp[OF I NF[unfolded R q] U]
  obtain ecap where ecap: "R_Q_U_ecap ?R ?Q ecap" 
    and U: "\<And> l r. (l,r) \<in> set U \<Longrightarrow> is_ur_closed_term' ?R (set U) ?Q ecap full_af (mv_xvar ` set (args l)) (mv_xvar r)"
    and t: "is_ur_closed_term' ?R (set U) ?Q ecap full_af (mv_xvar ` {s}) (mv_xvar (t))" 
    by auto
  from NFs have NF: "{s} \<cdot>set \<sigma> \<subseteq> NF_terms ?Q" unfolding q by auto
  have st: "(t \<cdot> \<sigma>, u) \<in> (qrstep ?nfs ?Q ?R)^*" 
    using steps unfolding R q nfs .
  have NFt: "\<sigma> ` vars_term t \<subseteq> NF_terms ?Q"
  proof (cases nfs)
    case True
    with nfsigma show ?thesis unfolding NF_subst_def vars_rule_def q by auto
  next
    case False
    show ?thesis
    proof (rule subsetI)
      fix u
      assume "u \<in> \<sigma> ` vars_term t"
      then obtain x where u: "u = \<sigma> x" and "x \<in> vars_term t" by auto
      with vars[OF False] have "x \<in> vars_term s" by auto
      from supteq_stable[OF supteq_Var[OF this], of \<sigma>] u have "s \<cdot> \<sigma> \<unrhd> u" by auto
      from NF_subterm[OF _ this] NF
      show "u \<in> NF_terms ?Q" unfolding q R by auto
    qed
  qed
  have st: "(t \<cdot> \<sigma>, u) \<in> (qrstep ?nfs ?Q (?R \<inter> set U))^*"
    by (rule R_Q_U_ecap.is_ur_closed_term[OF ecap st NFt NF U subset_refl _ t], insert varsR nfs q R, auto)
  have "qrstep ?nfs ?Q (?R \<inter> set U) \<subseteq> qrstep ?nfs ?Q (set U)" by auto
  from st rtrancl_mono[OF this]
  show ?thesis unfolding nfs q by auto
qed

definition inn_usable_rules_pair :: "('d, 'f :: key, string) dpp_ops
      \<Rightarrow> 'd \<Rightarrow> ('f, string)rule \<Rightarrow> ('f,string)rules"
  where "inn_usable_rules_pair I d \<equiv> 
  let inn = dpp_ops.NFQ_subset_NF_rules I d;
      R = dpp_ops.rules I d;
      qnf = dpp_ops.is_QNF I d;
      calc = get_ur_calc_impl_dpp I d full_af;
      nfs = dpp_ops.nfs I d;
      wwf = dpp_ops.wwf_rules I d;
      m   = dpp_ops.minimal I d
   in (\<lambda> (s,t). if inn \<and> (nfs \<or> (vars_term t \<subseteq> vars_term s)) \<and> (nfs \<or> m \<or> wwf) then
          calc (s,t) else R)"

lemma inn_usable_rules_pair:
  fixes I::"('d, 'f::{show,key}, string) dpp_ops" and d::"'d" 
  defines q: "q \<equiv> dpp_ops.Q I d"
      and R: "R \<equiv> dpp_ops.rules I d"
      and nfs: "nfs \<equiv> dpp_ops.nfs I d"
  assumes I: "dpp_spec I"
      and U: "U = inn_usable_rules_pair I d (s,t)"
      and steps: "(t \<cdot> \<sigma>, u) \<in> (qrstep nfs (set q) (set R))^*"
      and NF: "s \<cdot> \<sigma> \<in> NF_terms (set q)"      
      and nfsigma: "NF_subst nfs (s,t) \<sigma> (set q)"
      and SN: "dpp_ops.minimal I d \<Longrightarrow> SN_on (qrstep nfs (set q) (set R)) {t \<cdot> \<sigma>}"
  shows "(t \<cdot> \<sigma>, u) \<in> (qrstep nfs (set q) (set U))^*"
proof -
  interpret dpp_spec I by fact
  let ?R = "set (rules d)"
  let ?Q = "set (Q d)"
  let ?nfs = "NFS d"
  let ?m = "M d"
  let ?wwf = "wwf_rules d"
  let ?inn = "dpp_ops.NFQ_subset_NF_rules I d"
  show ?thesis 
  proof (cases "?inn \<and> (nfs \<or> vars_term t \<subseteq> vars_term s) \<and> (nfs \<or> ?m \<or> ?wwf)")
    case False
    hence "U = rules d" unfolding U inn_usable_rules_pair_def Let_def nfs split by auto
    thus ?thesis using steps unfolding R by auto
  next
    case True
    hence U: "U = get_ur_calc_impl_dpp I d full_af (s,t)" unfolding U inn_usable_rules_pair_def Let_def split nfs by auto 
    from True have inn: "NF_terms ?Q \<subseteq> NF_trs ?R" 
      and vars: "\<not> nfs \<Longrightarrow> vars_term t \<subseteq> vars_term s" by auto
    show ?thesis unfolding q nfs
      by (rule get_ur_calc_impl_dpp_steps[OF I U inn _ _ _ vars], insert nfsigma steps SN NF q R True nfs, auto)
  qed
qed

end
