(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Multiset_Extension_Impl
imports List2 Multiset_Extension2 List_Order_Impl
begin

(**
We use the lifting when it is sufficient to prove the non-strict
domination to also prove the strict one. For example if [a > x] then
it is sufficient to prove that [A \<ge> X] to get [A + {#a#} > X + {#x#}]
**)

lemma mul_ext_list_ext: "\<exists> s ns. list_order_extension_impl s ns mul_ext"
proof(intro exI)
  let ?s = "\<lambda> s ns. {(as,bs). (multiset_of as, multiset_of bs) \<in> s_mul_ext ns s}"
  let ?ns = "\<lambda> s ns. {(as,bs). (multiset_of as, multiset_of bs) \<in> ns_mul_ext ns s}" 
  let ?m = multiset_of
  show "list_order_extension_impl ?s ?ns mul_ext"
  proof
    fix s ns
    show "?s {(a,b). s a b} {(a,b). ns a b} = {(as,bs). fst (mul_ext (\<lambda> a b. (s a b, ns a b)) as bs)}" unfolding mul_ext_def Let_def by auto
  next
    fix s ns
    show "?ns {(a,b). s a b} {(a,b). ns a b} = {(as,bs). snd (mul_ext (\<lambda> a b. (s a b, ns a b)) as bs)}" unfolding mul_ext_def Let_def by auto
  next
    fix s ns s' ns' as bs
    assume "set as \<times> set bs \<inter> ns \<subseteq> ns'"
           "set as \<times> set bs \<inter> s \<subseteq> s'"
           "(as,bs) \<in> ?s s ns"
    thus "(as,bs) \<in> ?s s' ns'"
      using s_mul_ext_local_mono[of "?m as" "?m bs" ns ns' s s']
      unfolding set_of_multiset_of by auto
  next
    fix s ns s' ns' as bs
    assume "set as \<times> set bs \<inter> ns \<subseteq> ns'"
           "set as \<times> set bs \<inter> s \<subseteq> s'"
           "(as,bs) \<in> ?ns s ns"
    thus "(as,bs) \<in> ?ns s' ns'"
      using ns_mul_ext_local_mono[of "?m as" "?m bs" ns ns' s s']
      unfolding set_of_multiset_of by auto
  qed
qed


declare ns_mul_ext_def[simp]
declare s_mul_ext_def[simp]

primrec lift :: "(bool \<times> bool) \<Rightarrow> (bool \<times> bool)" where
  "lift (_, b) = (b,b)"

fun or2 :: "(bool \<times> bool) \<Rightarrow> (bool \<times> bool) \<Rightarrow> (bool \<times> bool)" where
  "or2 (b1, b2) (b3, b4) = (b1 \<or> b3, b2 \<or> b4)"

fun mul_ext_impl :: "('a \<Rightarrow> 'a \<Rightarrow> bool \<times> bool) \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> bool \<times> bool"
and ex_dominating :: "('a \<Rightarrow> 'a \<Rightarrow> bool \<times> bool) \<Rightarrow> 'a \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> bool \<times> bool"
where
  "mul_ext_impl f [] []       = (False, True)"              |
  "mul_ext_impl f [] _        = (False, False)"             |
  "mul_ext_impl f _  []       = (True, True)"               |
  "mul_ext_impl f xs (y # ys) = ex_dominating f y xs xs ys" |
  "ex_dominating f y []       _  _  = (False, False)" |
  "ex_dominating f y (b # bs) xs ys = (
     let (s, ns) = f b y in
     if s
     then or2 (lift (mul_ext_impl f (remove1 b xs) [y\<leftarrow>ys . \<not> fst (f b y)]))
              (ex_dominating f y bs xs ys)
     else if ns
          then (or2 (mul_ext_impl f (remove1 b xs) ys) (ex_dominating f y bs xs ys))
          else ex_dominating f y bs xs ys)"

lemma mul_ext_impl_sound_prelim:
  shows "mul_ext_impl f xs [] = mul_ext f xs []"
proof (cases xs)
case Nil
  thus ?thesis unfolding mul_ext_def
   using ns_mul_ext_bottom and s_mul_ext_bottom_strict
   by simp
next
case (Cons b bs) thus ?thesis unfolding mul_ext_def by simp
qed

lemma snd_or2[simp]: "snd (or2 b1 b2) = (snd b1 \<or> snd b2)"
  by (cases b1, cases b2, auto)

lemma or2_TT[simp]: "or2 (True, True) b = (True, True)"
by (cases b, simp)

lemma or2_FF[simp]: "or2 (False, False) b = b"
by (cases b, simp)

lemma lift_snd[simp]: "lift b = (snd b, snd b)"
by (cases b, simp)

lemma multiset_add:
  assumes "b \<in> set xs" and "fst (f b y)"
  and "snd (mul_ext f (remove1 b xs) (filter (\<lambda>y. \<not> fst (f b y)) ys))"
  shows "fst (mul_ext f xs ys)"
proof-
let ?s = "{(x,y). fst (f x y)}"
let ?ns = "{(x,y). snd (f x y)}"
let ?filt = "filter (\<lambda>y. \<not> fst (f b y)) ys"
let ?filt2 = "filter (\<lambda>y. fst (f b y)) ys"
let ?rem = "remove1 b xs"
let ?m = "multiset_of"
from assms(3) have "(?m ?rem, ?m ?filt) \<in> (ns_mul_ext ?ns ?s)"
 unfolding mul_ext_def by simp
then obtain ps qs P' Q' where
  P_part: "?m ?rem = ?m ps + P'" and
  Q_part: "?m ?filt = ?m qs + Q'" and
  pslqs: "length ps = length qs" and
  ns_ps_qs: "\<forall>i. i < length qs \<longrightarrow> (ps ! i, qs ! i) \<in> ?ns" and
  s_P'_Q': "\<forall>q. q \<in># Q' \<longrightarrow> (\<exists>p. p \<in># P' \<and> (p,q) \<in> ?s)"
unfolding ns_mul_ext_def by blast
from in_multiset_in_set and assms(1) have "b \<in># ?m xs" by fast
from multiset_of_remove1 and insert_DiffM2[OF this] have "?m xs = ?m ?rem + {#b#}" by simp
with P_part and union_assoc have Pb_part: "?m xs = ?m ps + (P' + {#b#})" by metis
from multiset_of_compl_union[of "\<lambda>y. fst (f b y)" ys, symmetric] and union_commute
 have "?m ys = ?m ?filt + ?m ?filt2" by metis
with union_assoc and Q_part
 have Qb_part: "?m ys = ?m qs + (Q' + ?m ?filt2)" by simp
{
  fix q assume q_in: "q \<in># (Q' + ?m ?filt2)"
  have "\<exists>p. p \<in># (P' + {#b#}) \<and> (p,q) \<in> ?s"
  proof (cases "q \<in># Q'")
  case True from s_P'_Q'[rule_format, OF this] show ?thesis by auto
  next
  case False
    with q_in have "q \<in># ?m ?filt2" by simp
    with in_multiset_in_set[of ?filt2 q] have "q \<in> set ?filt2" by simp
    with set_filter have "(b,q) \<in> ?s" by simp
    thus ?thesis by auto
  qed
}
hence s_P'b_Q'ys: "\<forall>q. q \<in># (Q' + ?m ?filt2) \<longrightarrow> (\<exists>p. p \<in># (P' + {#b#}) \<and> (p,q) \<in> ?s)" by simp
have "P' + {#b#} \<noteq> {#}" by simp
with Pb_part and Qb_part and pslqs and ns_ps_qs and s_P'b_Q'ys
have "(?m xs, ?m ys) \<in> (s_mul_ext ?ns ?s)" unfolding s_mul_ext_def by blast
thus ?thesis unfolding mul_ext_def by simp
qed

definition cond1 where
  "cond1 f bs y xs ys \<equiv>
  ((\<exists>b. b \<in> set bs \<and> fst (f b y) \<and> snd (mul_ext f (remove1 b xs) [y\<leftarrow>ys . \<not> fst (f b y)]))
  \<or> (\<exists>b. b \<in> set bs \<and> snd (f b y) \<and> fst (mul_ext f (remove1 b xs) ys)))"

lemma cond1_propagate:
  assumes "cond1 f bs y xs ys"
  shows "cond1 f (b # bs) y xs ys"
using assms unfolding cond1_def by auto

definition cond2 where
  "cond2 f bs y xs ys \<equiv> (cond1 f bs y xs ys
  \<or> (\<exists>b. b \<in> set bs \<and> snd (f b y) \<and> snd (mul_ext f (remove1 b xs) ys)))"

lemma cond2_propagate:
  assumes "cond2 f bs y xs ys"
  shows "cond2 f (b # bs) y xs ys"
using assms and cond1_propagate[of f bs y xs ys]
unfolding cond2_def by auto

lemma cond1_cond2:
  assumes "cond1 f bs y xs ys"
  shows "cond2 f bs y xs ys"
using assms unfolding cond2_def by simp

lemma ex_dominating_sound:
  assumes "\<forall>x. x \<in> set bs \<longrightarrow> x \<in> set xs"
  and "\<forall>as bs. length bs \<le> length ys \<longrightarrow> mul_ext_impl f as bs = mul_ext f as bs"
  shows "ex_dominating f y bs xs ys = (cond1 f bs y xs ys, cond2 f bs y xs ys)"
using assms(1) proof (induct bs)
case Nil show ?case unfolding cond1_def and cond2_def by simp
next
case (Cons b bs)
(** Introducing useful abbreviations. *)
  let ?m = "multiset_of"
  let ?s = "{(x,y). fst (f x y)}"
  let ?ns = "{(x,y). snd (f x y)}"
  let ?rem = "remove1 b xs"
  let ?filt = "filter (\<lambda>y. \<not> fst (f b y)) ys"
  let ?cond11 = "\<exists>b. b \<in> set bs \<and> fst (f b y) \<and>
   snd (mul_ext f (remove1 b xs) [y\<leftarrow>ys . \<not> fst (f b y)])"
  let ?cond12 = "\<exists>b. b \<in> set bs \<and> snd (f b y) \<and> fst (mul_ext f (remove1 b xs) ys)"
  let ?cond2 = "\<exists>b. b \<in> set bs \<and> snd (f b y) \<and> snd (mul_ext f (remove1 b xs) ys)"
(** Back to the proof. *)
  from Cons(2) have "\<forall>x. x \<in> set bs \<longrightarrow> x \<in> set xs" by simp
  note IH = Cons(1)[OF this]
  show ?case
  proof (cases "f b y")
  case (Pair s ns) note Psns = this
(**
    Given that there are a lot of computations involved, we chose to add
    intermediate [Hrew] results that give the current equation. This way
    we can easily see what is left to prove and we do not need to invoke
    all the intermediate result when we want to [show ?thesis].
**)
    hence Hrew: "ex_dominating f y (b # bs) xs ys = (
     if s
     then or2 (lift (mul_ext_impl f (remove1 b xs) (filter (\<lambda>y. \<not> fst (f b y)) ys)))
              (ex_dominating f y bs xs ys)
     else if ns
          then (or2 (mul_ext_impl f (remove1 b xs) ys) (ex_dominating f y bs xs ys))
          else ex_dominating f y bs xs ys)" by simp
    show ?thesis
    proof (cases s)
    case True
      with Hrew have Hrew: "ex_dominating f y (b # bs) xs ys =
        or2 (lift (mul_ext_impl f (remove1 b xs) (filter (\<lambda>y. \<not> fst (f b y)) ys)))
              (ex_dominating f y bs xs ys)" by simp
      from True and Psns have f_b_y: "fst (f b y)" by simp
      have "length ?filt \<le> length ys" by simp
      note IH2 = assms(2)[rule_format, OF this, of ?rem]
      show ?thesis
      proof (cases "mul_ext_impl f ?rem ?filt")
      case (Pair MIs MIns)
        with Hrew have Hrew: "ex_dominating f y (b # bs) xs ys =
         or2 (MIns, MIns) (ex_dominating f y bs xs ys)" by simp
        show ?thesis
        proof (cases MIns)
        case True
          with Hrew have Hrew: "ex_dominating f y (b # bs) xs ys = (True, True)" by simp
          from True and Pair and IH2 have HMns: "snd (mul_ext f ?rem ?filt)" by simp
          have b_in_bbs: "b \<in> set (b # bs)" by simp
          with HMns and f_b_y have "cond1 f (b # bs) y xs ys" unfolding cond1_def by fast
          with cond1_cond2 and Hrew show ?thesis by simp
        next
        case False
          with Hrew have Hrew: "ex_dominating f y (b # bs) xs ys =
            ex_dominating f y bs xs ys" by simp
          from False and Pair have HMIns: "\<not> snd (mul_ext_impl f ?rem ?filt)" by simp
          show ?thesis
          proof (cases "ex_dominating f y bs xs ys")
          case (Pair Ms Mns)
            with Hrew have Hrew: "ex_dominating f y (b # bs) xs ys = (Ms, Mns)" by simp
            show ?thesis
            proof (cases Ms)
            case True
              with Pair have HMs: "fst (ex_dominating f y bs xs ys)" by simp
              with IH and cond1_cond2 have HMns: "snd (ex_dominating f y bs xs ys)" by simp
              from Hrew and HMs and HMns and Pair
               have Hrew: "ex_dominating f y (b # bs) xs ys = (True, True)" by simp
              from HMs and IH and cond1_propagate have "cond1 f (b # bs) y xs ys" by simp
              thus ?thesis using cond1_cond2 and Hrew by simp
            next
            case False
              with Pair have HMs: "\<not> fst (ex_dominating f y bs xs ys)" by simp
              from False and Hrew
               have Hrew: "ex_dominating f y (b # bs) xs ys = (False, Mns)" by simp
              {
                assume "\<exists>c. c \<in> set (b # bs) \<and> fst (f c y) \<and>
                 snd (mul_ext f (remove1 c xs) [y\<leftarrow>ys . \<not> fst (f c y)])"
                then obtain a where
                  a_in_bbs: "a \<in> set (b # bs)" and
                  f_a_y: "fst (f a y)" and
                  ns_xs_ys: "snd (mul_ext f (remove1 a xs) [y\<leftarrow>ys . \<not> fst (f a y)])"
                by fast
                have "False"
                proof (cases "a = b")
                case True
                  have "length [y\<leftarrow>ys . \<not> fst (f a y)] \<le> length ys" by simp
                  with True and ns_xs_ys and HMIns and assms(2)[rule_format, OF this]
                   show ?thesis by auto
                next
                case False
                  with a_in_bbs have a_in_bs: "a \<in> set bs" by simp
                  with f_a_y and ns_xs_ys have "\<exists>c. c \<in> set bs \<and> fst (f c y) \<and>
                   snd (mul_ext f (remove1 c xs) [y\<leftarrow>ys . \<not> fst (f c y)])"
                  by fast
                  with IH have "fst (ex_dominating f y bs xs ys)"
                   unfolding cond1_def by simp
                  with HMs show ?thesis by auto
                qed
              }
              moreover
              {
                assume "\<exists>c. c \<in> set (b # bs) \<and> snd (f c y) \<and> fst (mul_ext f (remove1 c xs) ys)"
                then obtain a where
                  a_in_bbs: "a \<in> set (b # bs)" and
                  f_a_y: "snd (f a y)" and
                  s_xs_ys: "fst (mul_ext f (remove1 a xs) ys)"
                by fast
                hence H: "(?m (remove1 a xs), ?m ys) \<in> s_mul_ext ?ns ?s"
                 unfolding mul_ext_def by simp
                have "False"
                proof (cases "a = b")
                case True
                  from multiset_of_compl_union[of "\<lambda>y. fst (f a y)" ys]
                   and diff_union_cancelR[of "?m [y\<leftarrow>ys . \<not> fst (f a y)]"
                   "?m (filter (\<lambda>y. fst (f a y)) ys)"] and union_commute
                   have "?m [y\<leftarrow>ys . \<not> fst (f a y)] = ?m ys
                   - ?m (filter (\<lambda>y. fst (f a y)) ys)" by metis
                  with s_mul_ext_diff_multiset_r[OF H]
                   have "(?m (remove1 a xs), ?m [y\<leftarrow>ys . \<not> fst (f a y)]) \<in> s_mul_ext ?ns ?s"
                   by simp
                  with s_ns_mul_ext
                   have "snd (mul_ext f (remove1 a xs) [y\<leftarrow>ys . \<not> fst (f a y)])"
                   unfolding mul_ext_def by auto
                  with HMIns and True and IH2 show ?thesis by simp
                next
                case False
                  with a_in_bbs have a_in_bs: "a \<in> set bs" by simp
                  with f_a_y and s_xs_ys have "\<exists>c. c \<in> set bs \<and> snd (f c y) \<and>
                   fst (mul_ext f (remove1 c xs) ys)"
                  by fast
                  with IH have "fst (ex_dominating f y bs xs ys)"
                   unfolding cond1_def by simp
                  with HMs show ?thesis by auto
                qed
              }
              ultimately have Cond1: "\<not> (cond1 f (b # bs) y xs ys)"
               unfolding cond1_def by fast
              show ?thesis
              proof (cases Mns)
              case True
                with Hrew have Hrew: "ex_dominating f y (b # bs) xs ys = (False, True)"
                 by simp
                from True and Pair have HMns: "snd (ex_dominating f y bs xs ys)" by simp
                with IH and HMs and cond2_propagate have "cond2 f (b # bs) y xs ys" by simp
                with Hrew and Cond1 show ?thesis by simp
              next
              case False
                with Hrew have Hrew: "ex_dominating f y (b # bs) xs ys = (False, False)"
                 by simp
                from False and Pair have HMns: "\<not> snd (ex_dominating f y bs xs ys)"
                 by simp
                {
                  assume "cond2 f (b # bs) y xs ys"
                  with Cond1 obtain a where
                    a_in_bbs: "a \<in> set (b # bs)" and
                    f_a_y: "snd (f a y)" and
                    ns_xs_ys: "snd (mul_ext f (remove1 a xs) ys)"
                  unfolding cond2_def by auto
                  have False
                  proof (cases "a = b")
                  case False
                    with a_in_bbs have "a \<in> set bs" by simp
                    with f_a_y and ns_xs_ys and IH
                     have "snd (ex_dominating f y bs xs ys)"
                      unfolding cond2_def by auto
                    with HMns show ?thesis by fast
                  next
                  case True
                    from ns_xs_ys have H: "(?m (remove1 a xs), ?m ys) \<in> ns_mul_ext ?ns ?s"
                     unfolding mul_ext_def by simp
                    from multiset_of_compl_union[of "\<lambda>y. fst (f a y)" ys]
                     and diff_union_cancelR[of "?m [y\<leftarrow>ys . \<not> fst (f a y)]"
                      "?m (filter (\<lambda>y. fst (f a y)) ys)"] and union_commute
                     have "?m [y\<leftarrow>ys . \<not> fst (f a y)] = ?m ys
                      - ?m (filter (\<lambda>y. fst (f a y)) ys)" by metis
                    with ns_mul_ext_diff_multiset_r[OF H]
                     have "(?m (remove1 a xs), ?m [y\<leftarrow>ys . \<not> fst (f a y)]) \<in> ns_mul_ext ?ns ?s"
                     by simp
                    hence "snd (mul_ext f (remove1 a xs) [y\<leftarrow>ys . \<not> fst (f a y)])"
                     unfolding mul_ext_def by simp
                    with a_in_bbs and True and f_b_y have "cond1 f (b # bs) y xs ys"
                     unfolding cond1_def by fast
                    with Cond1 show ?thesis by fast
                  qed
                }
                hence Cond2: "\<not> (cond2 f (b # bs) y xs ys)" by fast
                from Cond1 and Cond2 and Hrew show ?thesis by simp
              qed
            qed    
          qed
        qed
      qed
    next
    case False note Hs = this
      with Hrew have Hrew: "ex_dominating f y (b # bs) xs ys = (
          if ns
          then (or2 (mul_ext_impl f (remove1 b xs) ys) (ex_dominating f y bs xs ys))
          else ex_dominating f y bs xs ys)" by simp
      show ?thesis
      proof (cases ns)
      case True
        with Hrew have Hrew: "ex_dominating f y (b # bs) xs ys =
         or2 (mul_ext_impl f ?rem ys) (ex_dominating f y bs xs ys)" by simp
        from True and Psns have f_b_y: "snd (f b y)" by simp
        have "length ys \<le> length ys" by simp
        note IH2 = assms(2)[rule_format, OF this, of ?rem]
        show ?thesis
        proof (cases "mul_ext_impl f (remove1 b xs) ys")
        case (Pair MIs MIns) note PMI = this
          with Hrew have Hrew: "ex_dominating f y (b # bs) xs ys =
           or2 (MIs, MIns) (ex_dominating f y bs xs ys)" by simp
          show ?thesis
          proof (cases MIs)
          case True
            with Pair have HMIs: "fst (mul_ext_impl f ?rem ys)" by simp
            with IH2 have HMIns: "snd (mul_ext_impl f ?rem ys)"
             unfolding mul_ext_def by auto
            from Hrew and Pair and HMIs and HMIns
             have Hrew: "ex_dominating f y (b # bs) xs ys = (True, True)" by simp
            from True and Pair have HMIs: "fst (mul_ext_impl f ?rem ys)" by simp
            with f_b_y and IH2 have "cond1 f (b # bs) y xs ys"
             unfolding cond1_def by auto
            with Hrew and cond1_cond2 show ?thesis by simp
          next
          case False
            with Pair have HMIs: "\<not> fst (mul_ext_impl f ?rem ys)" by simp
            from False and Hrew have Hrew: "ex_dominating f y (b # bs) xs ys =
              or2 (False, MIns) (ex_dominating f y bs xs ys)" by simp
            show ?thesis
            proof (cases "ex_dominating f y bs xs ys")
            case (Pair Ms Mns)
              with Hrew have Hrew: "ex_dominating f y (b # bs) xs ys = (Ms, MIns \<or> Mns)"
               by simp
              show ?thesis
              proof (cases Ms)
              case True
                with Pair and IH have Mns unfolding cond2_def by auto
                with True and Hrew
                 have Hrew: "ex_dominating f y (b # bs) xs ys = (True, True)" by simp
                from Pair and True and IH and cond1_propagate
                 have "cond1 f (b # bs) y xs ys" by auto
                with cond1_cond2 and Hrew show ?thesis by simp
              next
              case False
                with Hrew have Hrew: "ex_dominating f y (b # bs) xs ys =
                 (False, MIns \<or> Mns)" by simp
                from False and Pair have HMs: "\<not> fst (ex_dominating f y bs xs ys)" by simp
                {
                  assume "\<exists>c. c \<in> set (b # bs) \<and> fst (f c y) \<and>
                   snd (mul_ext f (remove1 c xs) [y\<leftarrow>ys . \<not> fst (f c y)])"
                  then obtain a where
                    a_in_bbs: "a \<in> set (b # bs)" and
                    f_a_y: "fst (f a y)" and
                    ns_xs_ys: "snd (mul_ext f (remove1 a xs) [y\<leftarrow>ys . \<not> fst (f a y)])"
                  by auto
                  have False
                  proof (cases "a = b")
                  case True
                    with f_a_y and Hs and Psns show ?thesis by simp
                  next
                  case False
                    with a_in_bbs have a_in_bs: "a \<in> set bs" by simp
                    with f_a_y and ns_xs_ys have "cond1 f bs y xs ys"
                     unfolding cond1_def by fast
                    with HMs and IH show ?thesis by simp
                  qed
                }
                moreover
                {
                  assume "\<exists>c. c \<in> set (b # bs) \<and> snd (f c y) \<and>
                   fst (mul_ext f (remove1 c xs) ys)"
                  then obtain a where
                    a_in_bbs: "a \<in> set (b # bs)" and
                    f_a_y: "snd (f a y)" and
                    s_xs_ys: "fst (mul_ext f (remove1 a xs) ys)"
                  by auto
                  have False
                  proof (cases "a = b")
                  case True with s_xs_ys and HMIs and IH2 show ?thesis by simp
                  next
                  case False
                    with a_in_bbs have a_in_bs: "a \<in> set bs" by simp
                    with f_a_y and s_xs_ys have "cond1 f bs y xs ys"
                     unfolding cond1_def by fast
                    with HMs and IH show ?thesis by simp
                  qed
                }
                ultimately have Cond1: "\<not> (cond1 f (b # bs) y xs ys)"
                 unfolding cond1_def by fast
                show ?thesis
                proof (cases MIns)
                case True
                  with Hrew have Hrew: "ex_dominating f y (b # bs) xs ys = (False, True)"
                   by simp
                  from PMI and True and IH2
                   have HMIns: "snd (mul_ext f (remove1 b xs) ys)" by simp
                  with f_b_y have Cond2: "cond2 f (b # bs) y xs ys"
                   unfolding cond2_def by auto
                  from Hrew and Cond1 and Cond2 show ?thesis by simp
                next
                case False
                  with Hrew have Hrew: "ex_dominating f y (b # bs) xs ys = (False, Mns)"
                   by simp
                  from False and PMI have HMIns: "\<not> snd (mul_ext_impl f ?rem ys)" by simp
                  show ?thesis
                  proof (cases Mns)
                  case True
                    with Hrew have Hrew: "ex_dominating f y (b # bs) xs ys =
                     (False, True)" by simp
                    from True and Pair have HMns: "snd (ex_dominating f y bs xs ys)" by simp
                    with cond2_propagate and IH and Hrew and Cond1 show ?thesis by simp
                  next
                  case False
                    with Hrew and Pair have Hrew: "ex_dominating f y (b # bs) xs ys =
                     (False, False)" by simp
                    from False and Pair have HMns: "\<not> snd (ex_dominating f y bs xs ys)"
                     by simp
                    {
                      assume "cond2 f (b # bs) y xs ys"
                      with Cond1 obtain a where
                        a_in_bbs: "a \<in> set (b # bs)" and
                        f_a_y: "snd (f a y)" and
                        ns_xs_ys: "snd (mul_ext f (remove1 a xs) ys)"
                      unfolding cond2_def by auto
                      have False
                      proof (cases "a = b")
                      case False
                        with a_in_bbs have "a \<in> set bs" by simp
                        with f_a_y and ns_xs_ys and IH
                        have "snd (ex_dominating f y bs xs ys)"
                         unfolding cond2_def by auto
                        with HMns show ?thesis by fast
                      next
                      case True
                        with ns_xs_ys and HMIns and IH2 show ?thesis by simp
                      qed
                    }
                    hence Cond2: "\<not> (cond2 f (b # bs) y xs ys)" by fast
                    from Hrew and Cond1 and Cond2 show ?thesis by simp
                  qed
                qed  
              qed
            qed
          qed
        qed
      next
      case False note Hns = this
        with Hrew have Hrew: "ex_dominating f y (b # bs) xs ys =
         ex_dominating f y bs xs ys" by simp
        show ?thesis
        proof (cases "ex_dominating f y bs xs ys")
        case (Pair Ms Mns)
          with Hrew have Hrew: "ex_dominating f y (b # bs) xs ys = (Ms, Mns)" by simp
          show ?thesis
          proof (cases Ms)
          case True
            with Pair have HMs: "fst (ex_dominating f y bs xs ys)" by simp
            with IH have HMs': "cond1 f bs y xs ys" by simp
            with cond1_propagate have Cond1: "cond1 f (b # bs) y xs ys" by fast
            with cond1_cond2 have Cond2: "cond2 f (b # bs) y xs ys" by fast
            from HMs' and Pair and IH and Hrew and cond1_cond2
             have Hrew: "ex_dominating f y (b # bs) xs ys = (True, True)" by auto
            from Hrew and Cond1 and Cond2 show ?thesis by simp
          next
          case False
            with Pair have HMs: "\<not> fst (ex_dominating f y bs xs ys)" by simp
            from False and Hrew have Hrew: "ex_dominating f y (b # bs) xs ys = (False, Mns)"
             by simp
            {
              assume  "\<exists>c. c \<in> set (b # bs) \<and> fst (f c y) \<and>
                 snd (mul_ext f (remove1 c xs) [y\<leftarrow>ys . \<not> fst (f c y)])"
                then obtain a where
                  a_in_bbs: "a \<in> set (b # bs)" and
                  f_a_y: "fst (f a y)" and
                  ns_xs_ys: "snd (mul_ext f (remove1 a xs) [y\<leftarrow>ys . \<not> fst (f a y)])"
                by fast
                have "False"
                proof (cases "a = b")
                case True with f_a_y and Hs and Psns show ?thesis by simp
                next
                case False
                  with a_in_bbs have a_in_bs: "a \<in> set bs" by simp
                  with f_a_y and ns_xs_ys have "\<exists>c. c \<in> set bs \<and> fst (f c y) \<and>
                   snd (mul_ext f (remove1 c xs) [y\<leftarrow>ys . \<not> fst (f c y)])"
                  by fast
                  with IH have "fst (ex_dominating f y bs xs ys)"
                   unfolding cond1_def by simp
                  with HMs show ?thesis by auto
                qed
            }
            moreover
            {
              assume "\<exists>c. c \<in> set (b # bs) \<and> snd (f c y) \<and> fst (mul_ext f (remove1 c xs) ys)"
              then obtain a where
                a_in_bbs: "a \<in> set (b # bs)" and
                f_a_y: "snd (f a y)" and
                s_xs_ys: "fst (mul_ext f (remove1 a xs) ys)"
              by fast
              hence H: "(?m (remove1 a xs), ?m ys) \<in> s_mul_ext ?ns ?s"
               unfolding mul_ext_def by simp
              have "False"
              proof (cases "a = b")
              case True with Hns and Psns and f_a_y show ?thesis by simp
              next
              case False
                with a_in_bbs have a_in_bs: "a \<in> set bs" by simp
                with f_a_y and s_xs_ys have "\<exists>c. c \<in> set bs \<and> snd (f c y) \<and>
                 fst (mul_ext f (remove1 c xs) ys)"
                by fast
                with IH have "fst (ex_dominating f y bs xs ys)"
                 unfolding cond1_def by simp
                with HMs show ?thesis by auto
              qed
            }
            ultimately have Cond1: "\<not> (cond1 f (b # bs) y xs ys)"
             unfolding cond1_def by fast
            show ?thesis
            proof (cases Mns)
            case True
              with Hrew have Hrew: "ex_dominating f y (b # bs) xs ys = (False, True)" by simp
              from True and Pair and IH have "cond2 f bs y xs ys" by simp
              with cond2_propagate have Cond2: "cond2 f (b # bs) y xs ys" by fast
              from Hrew and Cond1 and Cond2 show ?thesis by simp
            next
            case False
              with Hrew have Hrew: "ex_dominating f y (b # bs) xs ys = (False, False)" by simp
              from False and Pair and IH have HMns: "\<not> (cond2 f bs y xs ys)" by simp
              {
                assume "cond2 f (b # bs) y xs ys"
                with Cond1 obtain a where
                  a_in_bbs: "a \<in> set (b # bs)" and
                  f_a_y: "snd (f a y)" and
                  ns_xs_ys: "snd (mul_ext f (remove1 a xs) ys)"
                unfolding cond2_def by auto
                have False
                proof (cases "a = b")
                case True with f_a_y and Hns and Psns show ?thesis by simp
                next
                case False
                  with a_in_bbs have a_in_bs: "a \<in> set bs" by simp
                  with f_a_y and ns_xs_ys have "cond2 f bs y xs ys"
                   unfolding cond2_def by fast
                  with HMns show ?thesis by contradiction
                qed
              }
              hence Cond2: "\<not> (cond2 f (b # bs) y xs ys)" by fast
              from Hrew and Cond1 and Cond2 show ?thesis by simp
            qed
          qed
        qed
      qed
    qed
  qed
qed

lemma ex_dominating_is_mul_ext:
  assumes 
  "\<forall>as bs. length bs \<le> length ys \<longrightarrow> mul_ext_impl f as bs = mul_ext f as bs"
  shows "ex_dominating f y xs xs ys = mul_ext f xs (y # ys)"
proof-
let ?s = "{(x,y). fst (f x y)}"
let ?ns = "{(x,y). snd (f x y)}"
let ?m = "multiset_of"
have tauto: "\<forall>x. x \<in> set xs \<longrightarrow> x \<in> set xs" by simp
note Hrew = ex_dominating_sound[OF tauto assms(1), of y]
show ?thesis
proof (cases "\<exists>b. b \<in> set xs \<and> fst (f b y) \<and> snd (mul_ext f (remove1 b xs)
[y\<leftarrow>ys . \<not> fst (f b y)])")
case True
  hence "cond1 f xs y xs ys" unfolding cond1_def by simp
  with cond1_cond2 and Hrew have Hrew: "ex_dominating f y xs xs ys = (True, True)" by simp
  from True obtain a where
    a_in_xs: "a \<in> set xs" and
    f_a_y: "fst (f a y)" and
    ns_xs_ys: "snd (mul_ext f (remove1 a xs) [y\<leftarrow>ys . \<not> fst (f a y)])"
  by auto
  hence "[y\<leftarrow>ys . \<not> fst (f a y)]  = [z\<leftarrow>(y # ys) . \<not> fst (f a z)]" by simp
  with ns_xs_ys have H: "snd (mul_ext f (remove1 a xs) [z\<leftarrow>(y#ys) . \<not> fst (f a z)])" by simp
  with multiset_add and a_in_xs and f_a_y have Hs: "fst (mul_ext f xs (y # ys))" by fast
  with s_ns_mul_ext have Hns: "snd (mul_ext f xs (y # ys))" unfolding mul_ext_def by auto
  from Hrew and Hs and Hns and pair_collapse[of "mul_ext f xs (y # ys)"] show ?thesis by simp
next
case False note Cond11 = this
  show ?thesis
  proof (cases "\<exists>b. b \<in> set xs \<and> snd (f b y) \<and> fst (mul_ext f (remove1 b xs) ys)")
  case True
    hence "cond1 f xs y xs ys" unfolding cond1_def by simp
    with cond1_cond2 and Hrew have Hrew: "ex_dominating f y xs xs ys = (True, True)" by simp
    from True obtain a where
      a_in_xs: "a \<in> set xs" and
      f_a_y: "snd (f a y)" and
      s_xs_ys: "fst (mul_ext f (remove1 a xs) ys)"
    by auto
    from f_a_y have H: "\<forall>i. i < length [y] \<longrightarrow> ([a] ! i, [y] ! i) \<in> ?ns" by simp
    have "length [a] = length [y]" by simp
    from all_ns_ns_mul_ext[OF this H, of ?s]
     have ns_a_y: "({#a#}, {#y#}) \<in> (ns_mul_ext ?ns ?s)" by simp
    from s_xs_ys have s_xs_ys': "(?m (remove1 a xs), ?m ys) \<in> s_mul_ext ?ns ?s"
     unfolding mul_ext_def by simp
    from a_in_xs and in_multiset_in_set[of xs a] and insert_DiffM2
     and multiset_of_remove1[of a xs]
     have "?m (remove1 a xs) + {#a#} = ?m xs" by simp
    with s_ns_mul_ext_union_compat[OF s_xs_ys' ns_a_y]
     have "(?m xs, ?m (y # ys)) \<in> s_mul_ext ?ns ?s" by simp
    with s_ns_mul_ext[OF this] and Hrew show ?thesis unfolding mul_ext_def by simp
  next
  case False note Cond12 = this
    from Cond11 and Cond12 have Cond1: "\<not> (cond1 f xs y xs ys)" unfolding cond1_def by simp
    with Hrew have Hrew: "ex_dominating f y xs xs ys = (False, cond2 f xs y xs ys)" by simp
    {
      assume "fst (mul_ext f xs (y # ys))"
      then obtain as bs A' B' where
        A_part: "?m xs = ?m as + A'" and
        B_part: "?m (y # ys) = ?m bs + B'" and
        aslbs: "length as = length bs" and
        A'_ne: "A' \<noteq> {#}" and
        ns_as_bs: "\<forall>i. i < length bs \<longrightarrow> (as ! i, bs ! i) \<in> ?ns" and
        s_A'_B': "\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> ?s)"
      unfolding mul_ext_def by auto
      have y_in_yys: "y \<in># ?m (y # ys)" by simp
      {
        assume "y \<in># ?m bs"
        from in_multiset_of_idx[OF this] obtain i where
          i_bd: "i < length bs" and
          eq_y_bsi: "y = bs ! i"
        by auto
        with aslbs have i_bd2: "i < length as" by simp
        from ns_as_bs[rule_format, OF i_bd] and eq_y_bsi
         have f_a_y: "snd (f (as ! i) y)" by simp
        with nth_mem_multiset_of[OF i_bd2] and A_part
         and in_multiset_in_set[of xs "as ! i"] have asi_in_xs: "as ! i \<in> set xs" by simp
        from multiset_of_remove1[of "as ! i" xs] and remove_nth_soundness[OF i_bd2]
         and A_part and diff_union_swap_multiset[OF nth_mem_multiset_of[OF i_bd2], of A']
         and union_commute
         have A_part': "?m (remove1 (as ! i) xs) = ?m (remove_nth i as) + A'" by metis
        from remove_nth_soundness[OF i_bd] and B_part and union_commute and eq_y_bsi
         and diff_union_swap_multiset[OF nth_mem_multiset_of[OF i_bd], of B']
         and diff_union_cancelR[of "?m ys" "{#y#}"] and multiset_of.simps(2)[of y ys]
         have B_part': "?m ys = ?m (remove_nth i bs) + B'" by metis
        from remove_nth_length and i_bd and i_bd2 and aslbs
         have aislbis: "length (remove_nth i as) = length (remove_nth i bs)" by simp
        from A_part' and B_part' and aislbis and A'_ne and s_A'_B'
         and remove_nth_P_compat[OF aslbs[symmetric] ns_as_bs, of i]
         have "(?m (remove1 (as ! i) xs), ?m ys) \<in> s_mul_ext ?ns ?s"
         unfolding s_mul_ext_def by blast
        hence "fst (mul_ext f (remove1 (as ! i) xs) ys)" unfolding mul_ext_def by simp
        with asi_in_xs and f_a_y have "cond1 f xs y xs ys" unfolding cond1_def by fast
        with Cond1 have False by contradiction
      }
      moreover
      {
        assume y_in_B': "y \<in># B'"
        from s_A'_B'[rule_format, OF this] obtain a where
          a_in_A': "a \<in># A'" and
          f_a_y: "fst (f a y)"
        by auto
        let ?filt = "filter (\<lambda>y. fst (f a y)) (list_of B')"
        let ?nfilt = "[y\<leftarrow>(list_of B') . \<not> fst (f a y)]"
        let ?filt2 = "filter (\<lambda>y. fst (f a y)) ys"
        let ?nfilt2 = "filter (\<lambda>y. \<not> fst (f a y)) ys"
        from a_in_A' and A_part have a_in_mxs: "a \<in># multiset_of xs" by simp
        with in_multiset_in_set[of xs a] have a_in_xs: "a \<in> set xs" by simp
        from A_part and multiset_of_remove1 and diff_union_single_conv[OF a_in_A']
         have A_part': "?m (remove1 a xs) = ?m as + (A' - {#a#})" by simp
        from multiset_of_compl_union[of "\<lambda>y. \<not> fst (f a y)" "list_of B'"]
         and multiset_of_list_of[of B'] and diff_union_cancelR
        have H: "B' - ?m ?filt = ?m ?nfilt" by metis
        from multiset_of_compl_union[of "\<lambda>y. fst (f a y)" "list_of B'"]
         and multiset_of_list_of[of B'] and mset_le_add_left[of "?m ?filt"]
         have "?m ?filt \<le> B'" by metis
        with H and B_part and multiset_diff_union_assoc[OF this, of "?m bs"]
         have B_part': "?m (y # ys) - ?m ?filt = ?m bs + ?m ?nfilt" by metis
        {
          fix b assume b_in: "b \<in># ?m ?nfilt"
          hence "b \<in># ?m ?filt + ?m ?nfilt" by (simp add: multiset_of_filter)
          with multiset_of_list_of[of B'] and multiset_of_compl_union
           have b_in_B': "b \<in># B'" by simp
          from s_A'_B'[rule_format, OF b_in_B'] obtain ab where
            ab_in_A': "ab \<in># A'" and
            f_ab_b: "fst (f ab b)"
          by auto
          from b_in and in_multiset_in_set[of ?nfilt b] have Hs: "\<not> fst (f a b)" by simp
          have "\<exists>ab. ab \<in># (A' - {#a#}) \<and> (ab,b) \<in> ?s"
          proof (cases "ab = a")
          case True with Hs and f_ab_b show ?thesis by simp
          next
          case False
            with ab_in_A' have "ab \<in># (A' - {#a#})" by simp
            with f_ab_b show ?thesis by fast
          qed
        }
        hence s_A'a_B': "\<forall>b. b \<in># ?m ?nfilt \<longrightarrow> (\<exists>ab. ab \<in># (A' - {#a#}) \<and> (ab,b) \<in> ?s)" by simp
        from A_part' and B_part' and aslbs and ns_as_bs and s_A'a_B'
         have Hns: "(?m (remove1 a xs), ?m (y # ys) - ?m ?filt) \<in> ns_mul_ext ?ns ?s"
         unfolding ns_mul_ext_def by blast
        from f_a_y have S1: "?m  [y\<leftarrow>(y # ys) . \<not> fst (f a y)] = ?m  [y\<leftarrow>ys . \<not> fst (f a y)]"
         by simp
        from f_a_y have "?m ?nfilt2 = ?m [y\<leftarrow> (y # ys) . \<not> fst (f a y)]" by simp
        with B_part  and multiset_of_list_of[of B'] and multiset_of_filter and filter_union
         have EQ1: "?m ?nfilt2 = ?m [y\<leftarrow>bs . \<not> fst (f a y)] + ?m ?nfilt" by metis
        from multiset_of_compl_union[of "\<lambda>y. fst (f a y)" bs] and mset_le_add_left
         have H: "?m [y\<leftarrow>bs . fst (f a y)] \<le> ?m bs" by metis
        from multiset_of_compl_union[of "\<lambda>y. fst (f a y)" bs]
         and diff_union_cancelL[of "?m [y\<leftarrow>bs . fst (f a y)]"]
         have "?m [y\<leftarrow>bs . \<not> fst (f a y)] = ?m bs - ?m [y\<leftarrow>bs . fst (f a y)]" by metis
        with mset_le_multiset_union_diff_commute[OF H] and EQ1
         have "?m ?nfilt2 = ?m bs + ?m ?nfilt - ?m [y\<leftarrow>bs . fst (f a y)]" by simp
        with ns_mul_ext_diff_multiset_r[OF Hns] and B_part'
         have "(?m (remove1 a xs), ?m ?nfilt2) \<in> ns_mul_ext ?ns ?s" by simp
        hence "snd (mul_ext f (remove1 a xs) ?nfilt2)" unfolding mul_ext_def by simp
        with a_in_xs and f_a_y have "cond1 f xs y xs ys" unfolding cond1_def by fast
        with Cond1 have False by contradiction
      }
      ultimately have "False" using B_part and y_in_yys by auto
    }
    hence Hfst_mul_ext: "\<not> fst (mul_ext f xs (y # ys))" by fast
    show ?thesis
    proof (cases "\<exists>b. b \<in> set xs \<and> snd (f b y) \<and> snd (mul_ext f (remove1 b xs) ys)")
    case True
      hence Cond2: "cond2 f xs y xs ys" unfolding cond2_def by simp
      with Hrew have Hrew: "ex_dominating f y xs xs ys = (False, True)" by simp
      from True obtain a where
        a_in_xs: "a \<in> set xs" and
        f_a_y: "snd (f a y)" and
        ns_xs_ys: "snd (mul_ext f (remove1 a xs) ys)"
      by auto
      from ns_xs_ys obtain as bs A' B' where
        A_part: "?m (remove1 a xs) = ?m as + A'" and
        B_part: "?m ys = ?m bs + B'" and
        aslbs: "length as = length bs" and
        ns_as_bs: "\<forall>i. i < length bs \<longrightarrow> (as ! i, bs ! i) \<in> ?ns" and
        s_A'_B': "\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> ?s)"
      unfolding mul_ext_def by auto
      from f_a_y have H: "\<forall>i. i < length [y] \<longrightarrow> ([a] ! i, [y] ! i) \<in> ?ns" by simp
      have "length [a] = length [y]" by simp
      from all_ns_ns_mul_ext[OF this H, of ?s]
       have ns_a_y: "({#a#}, {#y#}) \<in> (ns_mul_ext ?ns ?s)" by simp
      from ns_xs_ys have s_xs_ys': "(?m (remove1 a xs), ?m ys) \<in> ns_mul_ext ?ns ?s"
       unfolding mul_ext_def by simp
      from a_in_xs and in_multiset_in_set[of xs a] and insert_DiffM2
       and multiset_of_remove1[of a xs]
       have "?m (remove1 a xs) + {#a#} = ?m xs" by simp
      with ns_ns_mul_ext_union_compat[OF s_xs_ys' ns_a_y]
       have "(?m xs, ?m (y # ys)) \<in> ns_mul_ext ?ns ?s" by simp
      with Hrew and Hfst_mul_ext show ?thesis unfolding mul_ext_def by simp
    next
    case False
      with Cond1 have Cond2: "\<not> (cond2 f xs y xs ys)" unfolding cond2_def by simp
      with Hrew have Hrew: "ex_dominating f y xs xs ys = (False, False)" by simp
      {
        assume "snd (mul_ext f xs (y # ys))"
        then obtain as bs A' B' where
          A_part: "?m xs = ?m as + A'" and
          B_part: "?m (y # ys) = ?m bs + B'" and
          aslbs: "length as = length bs" and
          ns_as_bs: "\<forall>i. i < length bs \<longrightarrow> (as ! i, bs ! i) \<in> ?ns" and
          s_A'_B': "\<forall>b. b \<in># B' \<longrightarrow> (\<exists>a. a \<in># A' \<and> (a,b) \<in> ?s)"
        unfolding mul_ext_def by auto
        have y_in_yys: "y \<in># ?m (y # ys)" by simp
        {
          assume "y \<in># ?m bs"
          from in_multiset_of_idx[OF this] obtain i where
            i_bd: "i < length bs" and
            eq_y_bsi: "y = bs ! i"
          by auto
          with aslbs have i_bd2: "i < length as" by simp
          from ns_as_bs[rule_format, OF i_bd] and eq_y_bsi
           have f_a_y: "snd (f (as ! i) y)" by simp
          with nth_mem_multiset_of[OF i_bd2] and A_part
           and in_multiset_in_set[of xs "as ! i"] have asi_in_xs: "as ! i \<in> set xs" by simp
          from multiset_of_remove1[of "as ! i" xs] and remove_nth_soundness[OF i_bd2]
           and A_part and diff_union_swap_multiset[OF nth_mem_multiset_of[OF i_bd2], of A']
           and union_commute
           have A_part': "?m (remove1 (as ! i) xs) = ?m (remove_nth i as) + A'" by metis
          from remove_nth_soundness[OF i_bd] and B_part and union_commute and eq_y_bsi
           and diff_union_swap_multiset[OF nth_mem_multiset_of[OF i_bd], of B']
           and diff_union_cancelR[of "?m ys" "{#y#}"] and multiset_of.simps(2)[of y ys]
           have B_part': "?m ys = ?m (remove_nth i bs) + B'" by metis
          from remove_nth_length and i_bd and i_bd2 and aslbs
           have aislbis: "length (remove_nth i as) = length (remove_nth i bs)" by simp
          from A_part' and B_part' and aislbis and s_A'_B'
           and remove_nth_P_compat[OF aslbs[symmetric] ns_as_bs, of i]
           have "(?m (remove1 (as ! i) xs), ?m ys) \<in> ns_mul_ext ?ns ?s"
           unfolding ns_mul_ext_def by blast
          hence "snd (mul_ext f (remove1 (as ! i) xs) ys)" unfolding mul_ext_def by simp
          with asi_in_xs and f_a_y have "cond2 f xs y xs ys" unfolding cond2_def by fast
          with Cond2 have False by contradiction
        }
        moreover
        {
          assume y_in_B': "y \<in># B'"
          from s_A'_B'[rule_format, OF this] obtain a where
            a_in_A': "a \<in># A'" and
            f_a_y: "fst (f a y)"
          by auto
          let ?filt = "filter (\<lambda>y. fst (f a y)) (list_of B')"
          let ?nfilt = "[y\<leftarrow>(list_of B') . \<not> fst (f a y)]"
          let ?filt2 = "filter (\<lambda>y. fst (f a y)) ys"
          let ?nfilt2 = "filter (\<lambda>y. \<not> fst (f a y)) ys"
          from a_in_A' and A_part have a_in_mxs: "a \<in># multiset_of xs" by simp
          with in_multiset_in_set[of xs a] have a_in_xs: "a \<in> set xs" by simp
          from A_part and multiset_of_remove1 and diff_union_single_conv[OF a_in_A']
           have A_part': "?m (remove1 a xs) = ?m as + (A' - {#a#})" by simp
          from multiset_of_compl_union[of "\<lambda>y. \<not> fst (f a y)" "list_of B'"]
           and multiset_of_list_of[of B'] and diff_union_cancelR
          have H: "B' - ?m ?filt = ?m ?nfilt" by metis
          from multiset_of_compl_union[of "\<lambda>y. fst (f a y)" "list_of B'"]
           and multiset_of_list_of[of B'] and mset_le_add_left[of "?m ?filt"]
           have "?m ?filt \<le> B'" by metis
          with H and B_part and multiset_diff_union_assoc[OF this, of "?m bs"]
           have B_part': "?m (y # ys) - ?m ?filt = ?m bs + ?m ?nfilt" by metis
          {
            fix b assume b_in: "b \<in># ?m ?nfilt"
            hence "b \<in># ?m ?filt + ?m ?nfilt" by (simp add: multiset_of_filter)
            with multiset_of_list_of[of B'] and multiset_of_compl_union
             have b_in_B': "b \<in># B'" by simp
            from s_A'_B'[rule_format, OF b_in_B'] obtain ab where
              ab_in_A': "ab \<in># A'" and
              f_ab_b: "fst (f ab b)"
            by auto
            from b_in and in_multiset_in_set[of ?nfilt b] have Hs: "\<not> fst (f a b)" by simp
            have "\<exists>ab. ab \<in># (A' - {#a#}) \<and> (ab,b) \<in> ?s"
            proof (cases "ab = a")
            case True with Hs and f_ab_b show ?thesis by simp
            next
            case False
              with ab_in_A' have "ab \<in># (A' - {#a#})" by simp
              with f_ab_b show ?thesis by fast
            qed
          }
          hence s_A'a_B': "\<forall>b. b \<in># ?m ?nfilt \<longrightarrow> (\<exists>ab. ab \<in># (A' - {#a#}) \<and> (ab,b) \<in> ?s)" by simp
          from A_part' and B_part' and aslbs and ns_as_bs and s_A'a_B'
           have Hns: "(?m (remove1 a xs), ?m (y # ys) - ?m ?filt) \<in> ns_mul_ext ?ns ?s"
           unfolding ns_mul_ext_def by blast
          from f_a_y have S1: "?m  [y\<leftarrow>(y # ys) . \<not> fst (f a y)] = ?m  [y\<leftarrow>ys . \<not> fst (f a y)]"
           by simp
          from f_a_y have "?m ?nfilt2 = ?m [y\<leftarrow> (y # ys) . \<not> fst (f a y)]" by simp
          with B_part  and multiset_of_list_of[of B'] and multiset_of_filter and filter_union
           have EQ1: "?m ?nfilt2 = ?m [y\<leftarrow>bs . \<not> fst (f a y)] + ?m ?nfilt" by metis
          from multiset_of_compl_union[of "\<lambda>y. fst (f a y)" bs] and mset_le_add_left
           have H: "?m [y\<leftarrow>bs . fst (f a y)] \<le> ?m bs" by metis
          from multiset_of_compl_union[of "\<lambda>y. fst (f a y)" bs]
           and diff_union_cancelL[of "?m [y\<leftarrow>bs . fst (f a y)]"]
           have "?m [y\<leftarrow>bs . \<not> fst (f a y)] = ?m bs - ?m [y\<leftarrow>bs . fst (f a y)]" by metis
          with mset_le_multiset_union_diff_commute[OF H] and EQ1
           have "?m ?nfilt2 = ?m bs + ?m ?nfilt - ?m [y\<leftarrow>bs . fst (f a y)]" by simp
          with ns_mul_ext_diff_multiset_r[OF Hns] and B_part'
           have "(?m (remove1 a xs), ?m ?nfilt2) \<in> ns_mul_ext ?ns ?s" by simp
          hence "snd (mul_ext f (remove1 a xs) ?nfilt2)" unfolding mul_ext_def by simp
          with a_in_xs and f_a_y have "cond1 f xs y xs ys" unfolding cond1_def by fast
          with Cond1 have False by contradiction
        }
        ultimately have "False" using B_part and y_in_yys by auto
      }
      hence "\<not> snd (mul_ext f xs (y # ys))" by fast
      with Hfst_mul_ext and pair_collapse
       have "mul_ext f xs (y # ys) = (False, False)" by (cases "mul_ext f xs (y # ys)", simp)
      with Hrew show ?thesis by simp
      qed
    qed
  qed
qed

lemma mul_ext_impl_sound_ind:
  assumes 
      "length ys \<le> n"
  shows "mul_ext_impl f xs ys = mul_ext f xs ys"
  using assms proof (induct n arbitrary: xs ys)
case 0
  hence ys_Nil: "ys = []" by simp
  with mul_ext_impl_sound_prelim show ?case by simp
next
case (Suc m)
  thus ?case proof (cases ys)
  case Nil with mul_ext_impl_sound_prelim show ?thesis by simp
  next
  case (Cons z zs) note Hys = this
    thus ?thesis proof (cases xs)
    case Nil
      from ns_mul_ext_bottom_uniqueness
       have "\<not> snd (mul_ext f [] (z # zs))"
       unfolding mul_ext_def by auto
      with s_mul_ext_bottom_strict and Cons and Nil
       show ?thesis unfolding mul_ext_def by simp
    next
    case (Cons b bs)
      hence Hrew: "mul_ext_impl f xs (z # zs) = ex_dominating f z xs xs zs" by simp
      {
        fix as::"'a list" and bs::"'a list"
        assume "length bs \<le> length zs"
        with Hys and Suc(2) have "length bs \<le> m" by simp
        with Suc(1) have "mul_ext_impl f as bs = mul_ext f as bs" by simp
      }
      hence "\<forall>as bs. length bs \<le> length zs \<longrightarrow> mul_ext_impl f as bs = mul_ext f as bs" by simp
      with ex_dominating_is_mul_ext[OF this] and Hrew and Hys
       show ?thesis by simp
    qed
  qed
qed

lemma mul_ext_impl_sound:
  shows "mul_ext_impl f xs ys = mul_ext f xs ys"
using mul_ext_impl_sound_ind by fast

lemma [code_unfold]: "mul_ext = mul_ext_impl"
  by (intro ext, unfold mul_ext_impl_sound, auto)

lemma mul_ext_impl_cong_prelim:
  assumes "\<And> x x'. x \<in> set xs \<Longrightarrow> x' \<in> set [] \<Longrightarrow> f x x' = g x x'"
  shows "mul_ext_impl f xs [] = mul_ext_impl g xs []"
using assms and mul_ext_impl_sound_prelim[of _ xs]
 and mul_ext_cong[of xs xs "[]" "[]"]
 by metis

lemma ex_dominating_cong:
  assumes "\<forall>x. x \<in> set bs \<longrightarrow> x \<in> set xs"
  and "\<And> x x'. x \<in> set xs \<Longrightarrow> x' \<in> set (y # ys) \<Longrightarrow> f x x' = g x x'"
  and "\<forall>as bs. length bs \<le> length ys \<longrightarrow>
   (\<forall>x x'. x \<in> set as \<longrightarrow> x' \<in> set bs \<longrightarrow> f x x' = g x x') \<longrightarrow>
   mul_ext_impl f as bs = mul_ext_impl g as bs"
  shows "ex_dominating f y bs xs ys = ex_dominating g y bs xs ys "
using assms(1) proof (induct bs)
case Nil show ?case by simp
next
case (Cons b bs)
  from Cons(2) have "\<forall>x. x \<in> set bs \<longrightarrow> x \<in> set xs" by simp
  note IH = Cons(1)[OF this]
  have Hrew: "\<And>f. ex_dominating f y (b # bs) xs ys = (
     let (s, ns) = f b y in
     if s
     then or2 (lift (mul_ext_impl f (remove1 b xs) [y\<leftarrow>ys . \<not> fst (f b y)]))
              (ex_dominating f y bs xs ys)
     else if ns
          then (or2 (mul_ext_impl f (remove1 b xs) ys) (ex_dominating f y bs xs ys))
          else ex_dominating f y bs xs ys)" by simp
  show ?case
  proof (cases "f b y")
  case (Pair s ns) note Psns = this
    from Cons(2) have b_in_xs: "b \<in> set xs" by simp
    from assms(2)[OF this] and Pair have Psns': "g b y = (s, ns)" by simp
    from Psns and Hrew have Hrewf: "ex_dominating f y (b # bs) xs ys = (
     if s
     then or2 (lift (mul_ext_impl f (remove1 b xs) [y\<leftarrow>ys . \<not> fst (f b y)]))
              (ex_dominating f y bs xs ys)
     else if ns
          then (or2 (mul_ext_impl f (remove1 b xs) ys) (ex_dominating f y bs xs ys))
          else ex_dominating f y bs xs ys)" by simp
    from Psns' and Hrew have Hrewg: "ex_dominating g y (b # bs) xs ys = (
     if s
     then or2 (lift (mul_ext_impl g (remove1 b xs) [y\<leftarrow>ys . \<not> fst (g b y)]))
              (ex_dominating g y bs xs ys)
     else if ns
          then (or2 (mul_ext_impl g (remove1 b xs) ys) (ex_dominating g y bs xs ys))
          else ex_dominating g y bs xs ys)" by simp
    from assms(2)[OF b_in_xs]
     have "\<forall>y. y \<in> set ys \<longrightarrow> ((\<not> fst (f b y)) = (\<not> fst (g b y)))" by simp
    with filter_cong[of ys ys "\<lambda>y. \<not> fst (f b y)"]
     have T: "[y\<leftarrow>ys . \<not> fst (f b y)] = [y\<leftarrow>ys . \<not> fst (g b y)]" by simp
    {
      fix x x'
      assume x_in_bxs:"x \<in> set (remove1 b xs)"
        and x'_in_bys: "x' \<in> set [y\<leftarrow>ys . \<not> fst (f b y)]"
      from x_in_bxs and set_mp[OF set_remove1_subset] have x_in_xs: "x \<in> set xs" by fast
      from x'_in_bys have "x' \<in> set (y # ys)" by simp
      from assms(2)[OF x_in_xs this] have "f x x' = g x x'" by assumption
    }
    hence "\<forall>x x'. x \<in> set (remove1 b xs) \<longrightarrow> x' \<in> set [y\<leftarrow>ys . \<not> fst (f b y)] \<longrightarrow>
     f x x' = g x x'" by fast
    with assms(3) have "mul_ext_impl f (remove1 b xs) [y\<leftarrow>ys . \<not> fst (f b y)] =
     mul_ext_impl g (remove1 b xs) [y\<leftarrow>ys . \<not> fst (f b y)]" by simp
    with T have Heq: "mul_ext_impl f (remove1 b xs) [y\<leftarrow>ys . \<not> fst (f b y)] =
     mul_ext_impl g (remove1 b xs) [y\<leftarrow>ys . \<not> fst (g b y)]" by simp
    {
      fix x x'
      assume x_in_bxs:"x \<in> set (remove1 b xs)"
        and x'_in_ys: "x' \<in> set ys"
      from x_in_bxs and set_mp[OF set_remove1_subset] have x_in_xs: "x \<in> set xs" by fast
      from x'_in_ys have "x' \<in> set (y # ys)" by simp
      from assms(2)[OF x_in_xs this] have "f x x' = g x x'" by assumption
    }
    hence "\<forall>x x'. x \<in> set (remove1 b xs) \<longrightarrow> x' \<in> set ys \<longrightarrow> f x x' = g x x'" by fast
    with assms(3) have "mul_ext_impl f (remove1 b xs) ys =
     mul_ext_impl g (remove1 b xs) ys" by simp
    with Heq and IH and Hrewf and Hrewg show ?thesis by simp
  qed
qed

lemma mul_ext_impl_cong_ind:
  assumes "\<And> x x'. x \<in> set xs \<Longrightarrow> x' \<in> set ys \<Longrightarrow> f x x' = g x x'"
  and "length ys \<le> n"
  shows "mul_ext_impl f xs ys = mul_ext_impl g xs ys"
using assms proof (induct n arbitrary: xs ys)
case (0 xs ys)
  hence "ys = []" by simp
  with 0(1) and mul_ext_impl_cong_prelim show ?case by metis
next
case (Suc n xs ys)
  show ?case
  proof (cases "ys")
  case Nil with Suc(2) and mul_ext_impl_cong_prelim show ?thesis by metis
  next
  case (Cons t ts) note Htts = this
    show ?thesis
    proof (cases xs)
    case Nil with Htts show ?thesis by simp
    next
    case (Cons s ss)
      with Htts have Hrew: "\<And>f. mul_ext_impl f xs ys = ex_dominating f t xs xs ts" by simp
      have xs_in_xs: "\<forall>x. x \<in> set xs \<longrightarrow> x \<in> set xs" by simp
      from Suc(2) and Htts
       have fg_ext: "\<And>x x'. \<lbrakk> x \<in> set xs ; x' \<in> set (t # ts) \<rbrakk> \<Longrightarrow> f x x' = g x x'" by simp
      {
        fix as::"'a list" and bs::"'a list"
        assume lbs_bd: "length bs \<le> length ts"
         and local_fg_ext: "\<forall>x x'. x \<in> set as \<longrightarrow> x' \<in> set bs \<longrightarrow> f x x' = g x x'"
        from Suc(3) and Htts and lbs_bd have "length bs \<le> n" by simp
        from Suc(1)[OF _ this] and local_fg_ext
        have "mul_ext_impl f as bs = mul_ext_impl g as bs" by simp
      }
      hence T: "\<forall>as bs. length bs \<le> length ts \<longrightarrow> (\<forall>x x'. x \<in> set as \<longrightarrow> x' \<in> set bs \<longrightarrow>
       f x x' = g x x') \<longrightarrow> mul_ext_impl f as bs = mul_ext_impl g as bs" by simp
      from fg_ext and ex_dominating_cong[OF xs_in_xs _ T]
       have "ex_dominating f t xs xs ts = ex_dominating g t xs xs ts" by simp
      thus ?thesis using Htts and Cons by simp
    qed
  qed
qed

lemma mul_ext_impl_cong[fundef_cong]:
  assumes "\<And> x x'. x \<in> set xs \<Longrightarrow> x' \<in> set ys \<Longrightarrow> f x x' = g x x'"
  shows "mul_ext_impl f xs ys = mul_ext_impl g xs ys"
using mul_ext_impl_cong_ind[OF assms(1)] by fast

declare ns_mul_ext_def[simp del]
declare s_mul_ext_def[simp del]

end
